# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for common display functions."""
from datetime import datetime
from typing import Any, Dict, Optional, List, Tuple, Set, Union

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype
from pandas.errors import OutOfBoundsDatetime
from bokeh.io import output_notebook, show
from bokeh.models.annotations import LegendItem
from bokeh.models import (
    ColumnDataSource,
    DatetimeTickFormatter,
    HoverTool,
    Label,
    LayoutDOM,
    Legend,
    RangeTool,
    Title,
)

# pylint: disable=too-many-lines
# pylint: disable=no-name-in-module
from bokeh.palettes import viridis

# pylint: enable=no-name-in-module
from bokeh.plotting import figure, reset_output
from bokeh.layouts import column

from .._version import VERSION
from ..common.exceptions import MsticpyParameterError
from ..common.utility import export, check_kwargs
from ..data.query_defns import ensure_df_datetimes

# pylint: disable=unused-import
# Importing to activate pandas accessors
from .timeline_pd_accessor import TimeLineAccessor  # noqa F401

# pylint: enable=unused-import

__version__ = VERSION
__author__ = "Ian Hellen"


# Constants
_WRAP = 50
_WRAP_CMDL = "WrapCmdl"
_DEFAULT_KWARGS = [
    "color",
    "data",
    "group_by",
    "height",
    "legend",
    "range_tool",
    "ref_event",
    "ref_time",
    "source_columns",
    "time_column",
    "title",
    "width",
    "yaxis",
    "ref_events",
    "ref_time_col",
    "ref_col",
    "ref_times",
]

_TL_KWARGS = [
    "alert",
    "overlay_color",
    "overlay_data",
    "ref_time",
    "ygrid",
    "xgrid",
    "hide",
]

_TIMELINE_HELP = (
    "https://msticpy.readthedocs.io/en/latest/msticpy.nbtools.html"
    "#msticpy.nbtools.timeline.{plot_type}"
)


@export
def display_timeline(
    data: Union[pd.DataFrame, dict],
    time_column: str = "TimeGenerated",
    source_columns: list = None,
    **kwargs,
) -> LayoutDOM:
    """
    Display a timeline of events.

    Parameters
    ----------
    data : Union[dict, pd.DataFrame]
        Either
        dict of data sets to plot on the timeline with the following structure::

            Key (str) - Name of data set to be displayed in legend
            Value (Dict[str, Any]) - containing:
                data (pd.DataFrame) - Data to plot
                time_column (str, optional) - Name of the timestamp column
                source_columns (list[str], optional) - source columns to use
                    in tooltips
                color (str, optional) - color of datapoints for this data
            If any of the last values are omitted, they default to the values
            supplied as parameters to the function (see below)

        Or
        DataFrame as a single data set or grouped into individual
        plot series using the `group_by` parameter
    time_column : str, optional
        Name of the timestamp column
        (the default is 'TimeGenerated')
    source_columns : list, optional
        List of default source columns to use in tooltips
        (the default is None)

    Other Parameters
    ----------------
    title : str, optional
        Title to display (the default is None)
    alert : SecurityAlert, optional
        Add a reference line/label using the alert time (the default is None)
    ref_event : Any, optional
        Add a reference line/label using the alert time (the default is None)
    ref_time : datetime, optional
        Add a reference line/label using `ref_time` (the default is None)
    group_by : str
        (where `data` is a DataFrame)
        The column to group timelines on
    legend: str, optional
        "left", "right", "inline" or "none"
        (the default is to show a legend when plotting multiple series
        and not to show one when plotting a single series)
    yaxis : bool, optional
        Whether to show the yaxis and labels (default is False)
    ygrid : bool, optional
        Whether to show the yaxis grid (default is False)
    xgrid : bool, optional
        Whether to show the xaxis grid (default is True)
    range_tool : bool, optional
        Show the the range slider tool (default is True)
    height : int, optional
        The height of the plot figure
        (the default is auto-calculated height)
    width : int, optional
        The width of the plot figure (the default is 900)
    color : str
        Default series color (default is "navy")
    overlay_data : pd.DataFrame:
        A second dataframe to plot as a different series.
    overlay_color : str
        Overlay series color (default is "green")
    hide : bool, optional
        If True, create but do not display the plot.
        By default, False.
    ref_events : pd.DataFrame, optional
        Add references line/label using the event times in the dataframe.
        (the default is None)
    ref_time_col : str, optional
        Add references line/label using the this column in `ref_events`
        for the time value (x-axis).
        (this defaults the value of the `time_column` parameter or 'TimeGenerated'
        `time_column` is None)
    ref_col : str, optional
        The column name to use for the label from `ref_events`
        (the default is None)
    ref_times : List[Tuple[datetime, str]], optional
        Add one or more reference line/label using (the default is None)

    Returns
    -------
    LayoutDOM
        The bokeh plot figure.

    """
    # Get args
    check_kwargs(kwargs, _DEFAULT_KWARGS + _TL_KWARGS)
    overlay_data: pd.DataFrame = kwargs.pop("overlay_data", None)
    overlay_columns: list = kwargs.pop("overlay_columns", source_columns)
    color: str = kwargs.get("color", "navy")  # don't remove this from kwargs
    overlay_color: str = kwargs.pop("overlay_color", "green")

    kwargs_sub = kwargs.copy()
    kwargs_sub["time_column"] = time_column
    kwargs_sub["source_columns"] = source_columns
    kwargs_sub["ref_time"], kwargs_sub["ref_label"] = _get_ref_event_time(**kwargs)

    if isinstance(data, pd.DataFrame):
        if overlay_data is not None:
            aggr_data = {
                "Primary": {
                    "data": data,
                    "time_column": time_column,
                    "source_columns": source_columns,
                    "color": color,
                },
                "Secondary": {
                    "data": overlay_data,
                    "time_column": time_column,
                    "source_columns": overlay_columns,
                    "color": overlay_color,
                },
            }
            return _display_timeline_dict(data=aggr_data, **kwargs_sub)

        # Create a dictionary from a grouped or simple series
        series_dict = _create_dict_from_grouping(
            data=data,
            source_columns=source_columns,
            time_column=time_column,
            group_by=kwargs.get("group_by", None),
            color=kwargs.get("color", "navy"),
        )
        return _display_timeline_dict(data=series_dict, **kwargs_sub)

    if isinstance(data, dict):
        return _display_timeline_dict(data, **kwargs_sub)
    return None


_TL_VALUE_KWARGS = ["kind", "y", "x"]


# pylint: disable=invalid-name, too-many-locals, too-many-statements, too-many-branches
@export  # noqa: C901, MC0001
def display_timeline_values(  # noqa: C901, MC0001
    data: pd.DataFrame,
    value_col: str = None,
    time_column: str = "TimeGenerated",
    source_columns: list = None,
    **kwargs,
) -> LayoutDOM:
    """
    Display a timeline of events.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame as a single data set or grouped into individual
        plot series using the `group_by` parameter
    time_column : str, optional
        Name of the timestamp column
        (the default is 'TimeGenerated')
    value_col : str
        The column name holding the value to plot vertically
    source_columns : list, optional
        List of default source columns to use in tooltips
        (the default is None)

    Other Parameters
    ----------------
    x : str, optional
        alias of `time_column`
    y : str
        an alias for `value_col`
    title : str, optional
        Title to display (the default is None)
    ref_event : Any, optional
        Add a reference line/label using the alert time (the default is None)
    ref_time : datetime, optional
        Add a reference line/label using `ref_time` (the default is None)
    ref_label : str, optional
        A label for the `ref_event` or `ref_time` reference item
    group_by : str
        (where `data` is a DataFrame)
        The column to group timelines on
    legend: str, optional
        "left", "right", "inline" or "none"
        (the default is to show a legend when plotting multiple series
        and not to show one when plotting a single series)
    yaxis : bool, optional
        Whether to show the yaxis and labels
    range_tool : bool, optional
        Show the the range slider tool (default is True)
    height : int, optional
        The height of the plot figure
        (the default is auto-calculated height)
    width : int, optional
        The width of the plot figure (the default is 900)
    color : str
        Default series color (default is "navy"). This is overridden by
        automatic color assignments if plotting a grouped chart
    kind : Union[str, List[str]]
        one or more glyph types to plot., optional
        Supported types are "circle", "line" and "vbar" (default is "vbar")
    hide : bool, optional
        If True, create but do not display the plot.
        By default, False.
    ref_events : pd.DataFrame, optional
        Add references line/label using the event times in the dataframe.
        (the default is None)
    ref_time_col : str, optional
        Add references line/label using the this column in `ref_events`
        for the time value (x-axis).
        (this defaults the value of the `time_column` parameter or 'TimeGenerated'
        `time_column` is None)
    ref_col : str, optional
        The column name to use for the label from `ref_events`
        (the default is None)
    ref_times : List[Tuple[datetime, str]], optional
        Add one or more reference line/label using (the default is None)

    Returns
    -------
    LayoutDOM
        The bokeh plot figure.

    """
    check_kwargs(kwargs, _DEFAULT_KWARGS + _TL_VALUE_KWARGS + ["y"])

    if data is None or not isinstance(data, pd.DataFrame) or data.empty:
        print("No data to plot.")
        return figure()

    reset_output()
    output_notebook()
    height: int = kwargs.pop("height", None)
    width: int = kwargs.pop("width", 900)
    title: str = kwargs.pop("title", None)
    time_column = kwargs.get("x", time_column)
    group_by: str = kwargs.get("group_by", None)
    show_yaxis: bool = kwargs.pop("yaxis", True)
    show_range: bool = kwargs.pop("range_tool", True)
    color: str = kwargs.get("color", "navy")
    legend_pos: str = kwargs.pop("legend", None)
    kind: Any = kwargs.pop("kind", ["vbar"])
    plot_kinds = kind if isinstance(kind, list) else [kind]
    hide: bool = kwargs.pop("hide", False)
    ref_events: pd.DataFrame = kwargs.pop("ref_events", None)
    ref_time_col: str = kwargs.pop("ref_time_col", time_column or "TimeGenerated")
    ref_col: str = kwargs.pop("ref_col", None)
    ref_times: List[Tuple[datetime, str]] = kwargs.pop("ref_times", None)
    value_col = value_col or kwargs.pop("y", None)
    if not value_col:
        raise ValueError("Must supply either `value_col` or `y` parameter.")

    ref_time, ref_label = _get_ref_event_time(**kwargs)

    if source_columns is None:
        source_columns = [value_col]
    if value_col not in source_columns:
        source_columns.append(value_col)
    check_df_columns(
        data, source_columns + [time_column], _TIMELINE_HELP, "display_timeline_values"
    )
    data = ensure_df_datetimes(data=data, columns=time_column)
    graph_df, group_count_df, tool_tip_columns, series_count = _create_data_grouping(
        data, source_columns, time_column, group_by, color
    )

    tooltips, formatters = _create_tool_tips(data, tool_tip_columns)
    hover = HoverTool(tooltips=tooltips, formatters=formatters)

    # Create the Plot figure
    title = title or "Timeline"
    min_time = graph_df[time_column].min()
    max_time = graph_df[time_column].max()
    start_range = min_time - ((max_time - min_time) * 0.1)
    end_range = max_time + ((max_time - min_time) * 0.1)
    height = height or _calc_auto_plot_height(series_count)

    plot = figure(
        x_range=(start_range, end_range),
        min_border_left=50,
        plot_height=height,
        plot_width=width,
        x_axis_label="Event Time",
        x_axis_type="datetime",
        x_minor_ticks=10,
        y_axis_label=value_col,
        tools=[hover, "xwheel_zoom", "box_zoom", "reset", "save", "xpan"],
        toolbar_location="above",
        title=title,
    )
    plot.yaxis.visible = show_yaxis
    plot.ygrid.minor_grid_line_color = "navy"
    plot.ygrid.minor_grid_line_alpha = 0.1
    plot.ygrid.grid_line_color = "navy"
    plot.ygrid.grid_line_alpha = 0.3
    plot.xgrid.minor_grid_line_color = "navy"
    plot.xgrid.minor_grid_line_alpha = 0.1
    plot.xgrid.grid_line_color = "navy"
    plot.xgrid.grid_line_alpha = 0.3

    # set the tick datetime formatter
    plot.xaxis[0].formatter = _get_tick_formatter()

    # plot groups individually so that we can create an interactive legend
    if group_by:
        legend_items = []
        for _, group_id in group_count_df[group_by].items():
            first_group_item = graph_df[graph_df[group_by] == group_id].iloc[0]
            legend_label = str(first_group_item[group_by])
            inline_legend = str(group_id)
            group_color = first_group_item["color"]
            row_source = ColumnDataSource(graph_df[graph_df[group_by] == group_id])
            p_series = []
            # create default plot args
            plot_args: Dict[str, Any] = dict(
                x=time_column, alpha=0.7, source=row_source
            )
            if legend_pos != "none":
                plot_args["legend_label"] = str(inline_legend)

            if "vbar" in plot_kinds:
                p_series.append(
                    plot.vbar(top=value_col, width=4, color="color", **plot_args)
                )
            if "circle" in plot_kinds:
                p_series.append(
                    plot.circle(y=value_col, size=4, color="color", **plot_args)
                )
            if "line" in plot_kinds:
                p_series.append(
                    plot.line(
                        y=value_col, line_width=2, line_color=group_color, **plot_args
                    )
                )
            if not inline_legend:
                legend_items.append((legend_label, p_series))

        if legend_pos == "inline":
            # Position the inline legend
            plot.legend.location = "top_left"
            plot.legend.click_policy = "hide"
        elif legend_pos in {"left", "right"}:
            # Create the legend box outside of the plot area
            ext_legend = Legend(
                items=legend_items,
                location="center",
                click_policy="hide",
                label_text_font_size="8pt",
            )
            plot.add_layout(ext_legend, legend_pos)
    else:
        plot_args = dict(
            x=time_column, color=color, alpha=0.7, source=ColumnDataSource(graph_df)
        )
        if "vbar" in plot_kinds:
            plot.vbar(top=value_col, width=4, **plot_args)
        if "circle" in plot_kinds:
            plot.circle(y=value_col, size=4, **plot_args)
        if "line" in plot_kinds:
            plot.line(y=value_col, line_width=2, **plot_args)

    # if we have a reference, plot the time as a line
    if ref_time is not None:
        _add_ref_line(plot, ref_time, ref_label, data[value_col].max())
    elif ref_events is not None or ref_times is not None:
        _plot_ref_events(
            plot=plot,
            ref_events=ref_events,
            time_col=ref_time_col,
            group_count=series_count,
            ref_col=ref_col,
            ref_times=ref_times,
        )

    if show_range:
        rng_select = _create_range_tool(
            data=graph_df,
            min_time=min_time,
            max_time=max_time,
            plot_range=plot.x_range,
            width=width,
            height=height,
            time_column=time_column,
        )
        plot_layout = column(plot, rng_select)
    else:
        plot_layout = plot

    if not hide:
        show(plot_layout)
    return plot_layout


# pylint: enable=invalid-name,too-many-locals, too-many-statements, too-many-branches


# pylint: disable=too-many-locals, too-many-statements, too-many-branches
def _display_timeline_dict(data: dict, **kwargs) -> figure:  # noqa: C901, MC0001
    """
    Display a timeline of events.

    Parameters
    ----------
    data : dict
        Data points to plot on the timeline.
            Need to contain:
                Key - Name of data type to be displayed in legend
                Value - dict of data containing:
                    data : pd.DataFrame
                        Data to plot
                    time_column : str
                        Name of the timestamp column
                    source_columns : list
                        List of source columns to use in tooltips
                    color: str
                        Color of datapoints for this data
    Other Parameters
    ----------------
    ref_time : datetime, optional
        Input reference line to display (the default is None)
    title : str, optional
        Title to display (the default is None)
    time_column : str, optional
        Name of the timestamp column
        (the default is 'TimeGenerated')
    legend: str, optional
        Where to position the legend
        None, left, right or inline (default is None)
    yaxis : bool, optional
        Whether to show the yaxis and labels
    range_tool : bool, optional
        Show the the range slider tool (default is True)
    source_columns : list, optional
        List of default source columns to use in tooltips
        (the default is None)
    height : int, optional
        The height of the plot figure
        (the default is auto-calculated height)
    width : int, optional
        The width of the plot figure (the default is 900)
    ref_events : pd.DataFrame, optional
        Add references line/label using the event times in the dataframe.
        (the default is None)
    ref_time_col : str, optional
        Add references line/label using the this column in `ref_events`
        for the time value (x-axis).
        (this defaults the value of the `time_column` parameter or 'TimeGenerated'
        `time_column` is None)
    ref_col : str, optional
        The column name to use for the label from `ref_events`
        (the default is None)
    ref_times : List[Tuple[datetime, str]], optional
        Add one or more reference line/label using (the default is None)

    Returns
    -------
    figure
        The bokeh plot figure.

    """
    reset_output()
    output_notebook()

    height: int = kwargs.pop("height", None)
    width: int = kwargs.pop("width", 900)
    ref_time: Any = kwargs.pop("ref_time", None)
    ref_label: str = kwargs.pop("ref_label", None)
    title: str = kwargs.pop("title", None)
    legend_pos: str = kwargs.pop("legend", None)
    show_yaxis: bool = kwargs.pop("yaxis", False)
    show_range: bool = kwargs.pop("range_tool", True)
    xgrid: bool = kwargs.pop("xgrid", True)
    ygrid: bool = kwargs.pop("ygrid", False)
    hide: bool = kwargs.pop("hide", False)
    ref_events: pd.DataFrame = kwargs.pop("ref_events", None)
    ref_col: str = kwargs.pop("ref_col", None)
    ref_time_col: str = kwargs.pop(
        "ref_time_col", kwargs.get("time_column", "TimeGenerated")
    )
    ref_times: List[Tuple[datetime, str]] = kwargs.pop("ref_times", None)

    tool_tip_columns, min_time, max_time = _unpack_data_series_dict(data, **kwargs)
    series_count = len(data)

    tooltips, formatters = _create_tool_tips(data, tool_tip_columns)
    hover = HoverTool(tooltips=tooltips, formatters=formatters)

    title = f"Timeline: {title}" if title else "Event Timeline"

    start_range, end_range, min_time, max_time = _get_time_bounds(min_time, max_time)
    height = height or _calc_auto_plot_height(len(data))
    y_range = ((-1 / series_count), series_count - 1 + (1 / series_count))
    plot = figure(
        x_range=(start_range, end_range),
        y_range=y_range,
        min_border_left=50,
        plot_height=height,
        plot_width=width,
        x_axis_label="Event Time",
        x_axis_type="datetime",
        x_minor_ticks=10,
        tools=[hover, "xwheel_zoom", "box_zoom", "reset", "save", "xpan"],
        title=title,
    )

    _set_axes_and_grids(data, plot, show_yaxis, ygrid, xgrid)

    # Create plot bar to act as as range selector
    rng_select = _create_range_tool(
        data=data,
        min_time=min_time,
        max_time=max_time,
        plot_range=plot.x_range,
        width=width,
        height=height,
    )

    # set the tick datetime formatter
    plot.xaxis[0].formatter = _get_tick_formatter()
    # plot the data
    _plot_series(data, plot, legend_pos)

    if ref_time is not None:
        _add_ref_line(plot, ref_time, ref_label, len(data))
    elif ref_events is not None or ref_times is not None:
        _plot_ref_events(
            plot=plot,
            ref_events=ref_events,
            time_col=ref_time_col,
            group_count=series_count,
            ref_col=ref_col,
            ref_times=ref_times,
        )

    plot_layout = column(plot, rng_select) if show_range else plot
    if not hide:
        show(plot_layout)

    return plot_layout


# pylint: enable=too-many-locals, too-many-statements, too-many-branches


def _plot_series(data, plot, legend_pos):
    """Plot data series and add legend."""
    # plot groups individually so that we can create an interactive legend
    # if legend_pos is "inline", we add add the normal legend inside the plot
    # if legend_pos is "left" or "right", we add the legend to the side
    if len(data) > 1 and not legend_pos:
        legend_pos = "left"
    legend_items = []
    for ser_name, series_def in data.items():

        if legend_pos == "inline":
            p_series = plot.diamond(
                x=series_def["time_column"],
                y="y_index",
                color=series_def["color"],
                alpha=0.5,
                size=10,
                source=series_def["source"],
                legend_label=str(ser_name),
            )
        else:
            p_series = plot.diamond(
                x=series_def["time_column"],
                y="y_index",
                color=series_def["color"],
                alpha=0.5,
                size=10,
                source=series_def["source"],
            )
        if legend_pos in ["left", "right"]:
            legend_items.append(
                LegendItem(
                    label=str(ser_name),
                    renderers=[p_series],
                )
            )

    if legend_pos == "inline":
        # Position the inline legend
        plot.legend.location = "center_left"
        plot.legend.click_policy = "hide"
    elif legend_pos in ["left", "right"]:
        # Create the legend box outside of the plot area
        ext_legend = Legend(
            items=legend_items[::-1],  # the legend is in the wrong order otherwise
            location="center",
            click_policy="hide",
            label_text_font_size="8pt",
        )
        plot.add_layout(ext_legend, legend_pos)


def check_df_columns(
    data: pd.DataFrame, req_columns: List[str], help_uri: str, plot_type: str
):
    """
    Check that specified columns are in the DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        [description]
    req_columns : List[str]
        [description]
    help_uri : str
        [description]
    plot_type : str
        [description]

    Raises
    ------
    MsticpyParameterError
        If one or more columns not found in `data`

    """
    missing_cols = set(req_columns) - set(data.columns)
    if missing_cols:
        raise MsticpyParameterError(
            title="Columns not found in DataFrame",
            help_uri=help_uri.format(plot_type=plot_type),
            parameter=missing_cols,
        )


def _set_axes_and_grids(data, plot, show_yaxis, ygrid, xgrid):
    """Set the axes visibility and grids according to parameters."""
    plot.yaxis.visible = show_yaxis
    if show_yaxis and data:
        y_labels = {ser_def["y_index"]: str(lbl) for lbl, ser_def in data.items()}
        plot.yaxis.major_label_overrides = y_labels
    if ygrid:
        plot.ygrid.minor_grid_line_color = "navy"
        plot.ygrid.minor_grid_line_alpha = 0.1
        plot.ygrid.grid_line_color = "navy"
        plot.ygrid.grid_line_alpha = 0.3
    else:
        plot.ygrid.grid_line_color = None
    if xgrid:
        plot.xgrid.minor_grid_line_color = "navy"
        plot.xgrid.minor_grid_line_alpha = 0.3
    else:
        plot.xgrid.grid_line_color = None


def _get_time_bounds(min_time, max_time):
    """Return start and end range, coping with out-of-bounds error."""
    try:
        start_range = min_time - ((max_time - min_time) * 0.1)
        end_range = max_time + ((max_time - min_time) * 0.1)
    except OutOfBoundsDatetime:
        min_time = min_time.to_pydatetime()
        max_time = max_time.to_pydatetime()
        start_range = min_time - ((max_time - min_time) * 0.1)
        end_range = max_time + ((max_time - min_time) * 0.1)
    return start_range, end_range, min_time, max_time


# pylint: disable=too-many-locals
def _unpack_data_series_dict(data, **kwargs):
    time_column: str = kwargs.pop("time_column", "TimeGenerated")
    source_columns: list = kwargs.pop("source_columns", None)

    # Process the input dictionary
    # Take each item that is passed and fill in blanks and add a y_index
    tool_tip_columns: Set[str] = set()
    min_time = None
    max_time = None
    y_index = 0

    # Create a color map in case colors have not been specified
    # (Shift the Viridis palatte so we lose the top, harder-to-see colors)
    series_count = len(data)
    colors, palette_size = _get_color_palette(series_count)

    for ser_name, series_def in data.items():
        data_columns: Set[str] = set()
        series_data = series_def["data"]

        if (
            series_data is None
            or not isinstance(series_data, pd.DataFrame)
            or series_data.empty
        ):
            print(f"No data to plot for series {ser_name}.")
            continue

        # if the series has source columns, use those
        # or fall back to global source cols or defaults
        series_def_src_cols = _get_def_source_cols(series_data, source_columns)
        src_cols = series_def.get("source_columns", series_def_src_cols)
        data_columns.update(src_cols or series_def_src_cols)

        time_col = series_def.get("time_column", None)
        if not time_col:
            time_col = time_column
            series_def["time_column"] = time_col
        # ensure named time columns are in datetime format
        series_data = ensure_df_datetimes(data=series_data, columns=time_col)
        data_columns.update([time_col])
        # add the data columns to the tool tip column set
        tool_tip_columns.update(data_columns)

        # Create the Column data source to plot
        graph_df = series_data[list(data_columns)].copy()
        graph_df["y_index"] = y_index
        series_def["y_index"] = y_index
        ser_color = series_def.get("color", None)
        if not ser_color:
            ser_color = colors[y_index % palette_size]
            series_def["color"] = ser_color
        # Wrap tooltip lines longer than 50 chars
        _wrap_df_columns(graph_df, 50)
        series_def["source"] = ColumnDataSource(graph_df)
        y_index += 1

        # calculate min/max time from this set
        if min_time is None:
            min_time = series_data[time_col].min()
            max_time = series_data[time_col].max()
        else:
            min_time = min(min_time, series_data[time_col].min())
            max_time = max(max_time, series_data[time_col].max())

    return tool_tip_columns, min_time, max_time


# pylint: enable=too-many-locals


def _get_def_source_cols(data, source_columns):
    if not source_columns:
        data_columns = set()
        if all(
            col in data.columns for col in ["NewProcessName", "EventID", "CommandLine"]
        ):
            data_columns = set(["NewProcessName", "EventID", "CommandLine"])
    else:
        data_columns = set(source_columns)
    return data_columns


def _create_data_grouping(data, source_columns, time_column, group_by, color):
    data_columns = _get_def_source_cols(data, source_columns)
    # If the time column not explicitly specified in source_columns, add it
    data_columns.add(time_column)
    tool_tip_columns = data_columns.copy()
    # create group frame so that we can color each group separately
    if group_by:
        group_count_df = (
            data[[group_by, time_column]]
            .groupby(group_by)
            .count()
            .reset_index()
            .rename(columns={time_column: "count"})
        )
        group_count_df["y_index"] = group_count_df.index

        # Shift the Viridis palatte so we lose the top, harder-to-see colors
        series_count = len(group_count_df)
        colors, palette_size = _get_color_palette(series_count)
        group_count_df["color"] = group_count_df.apply(
            lambda x: colors[x.y_index % palette_size], axis=1
        )
        # re-join with the original data
        data_columns.update([group_by, "y_index", "color"])
        clean_data = data.drop(columns=["y_index", "color"], errors="ignore")
        graph_df = clean_data.merge(group_count_df, on=group_by)[list(data_columns)]
    else:
        graph_df = data[list(data_columns)].copy()
        graph_df["color"] = color
        graph_df["y_index"] = 1
        series_count = 1
        group_count_df = None
    return graph_df, group_count_df, tool_tip_columns, series_count


# pylint: enable=too-many-arguments


def _create_dict_from_grouping(data, source_columns, time_column, group_by, color):
    data_columns = _get_def_source_cols(data, source_columns)
    # If the time column not explicitly specified in source_columns, add it
    data_columns.add(time_column)

    series_dict: Dict[str, Dict] = {}
    # create group frame so that we can color each group separately
    if group_by:
        data_columns.add(group_by)
        grouped_data = data[list(data_columns)].groupby(group_by)

        series_count = len(grouped_data)
        colors, palette_size = _get_color_palette(series_count)
        for color_index, (group_name, group_df) in enumerate(grouped_data):
            series_dict[str(group_name)] = dict(
                data=group_df,
                time_column=time_column,
                source_columns=source_columns,
                color=colors[color_index % palette_size],
            )
    else:
        group_df = data[list(data_columns)].copy()
        series_dict["unnamed series"] = dict(
            data=group_df,
            time_column=time_column,
            source_columns=source_columns,
            color=color,
        )

    return series_dict


def _get_ref_event_time(**kwargs) -> Tuple[Optional[Any], Union[Any, str]]:
    """Extract the reference time from kwargs."""
    ref_alert = kwargs.get("alert", None)
    if ref_alert is not None:
        ref_event = ref_alert
        ref_label = "Alert time"
    else:
        ref_event = kwargs.get("ref_event", None)
        ref_label = "Event time"

    if ref_event is not None:
        if isinstance(ref_event, pd.DataFrame):
            ref_event = ref_event.iloc[0]
        ref_time = getattr(ref_event, "StartTimeUtc", None)
        if not ref_time:
            ref_time = getattr(ref_event, "TimeGenerated", None)
    else:
        ref_time = kwargs.get("ref_time", None)
        ref_label = "Ref time"
    return ref_time, kwargs.get("ref_label", ref_label)  # type: ignore


def _get_datetime_tooltip(col: str, dataset: pd.DataFrame):
    """Return tooltip and formatter entries for column."""
    if " " in col:
        disp_col = col.replace(" ", "_")
        tt_col = f"{{{col}}}"
    else:
        disp_col = tt_col = col
    if col in dataset and is_datetime64_any_dtype(dataset[col]):
        col_tooltip = f"@{tt_col}{{%F %T.%3N}}"
        col_fmt: Dict[Any, Any] = {f"@{tt_col}": "datetime"}
    else:
        col_tooltip = f"@{tt_col}"
        col_fmt = {}
    return disp_col, col_tooltip, col_fmt


def _create_tool_tips(
    data: Union[pd.DataFrame, Dict[str, pd.DataFrame]], columns: List[str]
) -> Tuple[List[Tuple[str, str]], Dict[str, str]]:
    """Create formatting for tool tip columns."""
    formatters: Dict[str, str] = {}
    # if this is a dict we need to unpack each dataframe and process
    # the tooltip columns for all of the data sets.
    if isinstance(data, dict):
        tool_tip_dict = {}
        for data_set in data.values():
            data_df = data_set.get("data", {})
            for col in columns:
                disp_col, col_tooltip, col_fmt = _get_datetime_tooltip(col, data_df)
                tool_tip_dict[disp_col] = col_tooltip
                formatters.update(col_fmt)
        return list(tool_tip_dict.items()), formatters

    # If just a dataframe we just process the columns against this
    tool_tip_items = []
    for col in columns:
        disp_col, col_tooltip, col_fmt = _get_datetime_tooltip(col, data)
        tool_tip_items.append((disp_col, col_tooltip))
        formatters.update(col_fmt)

    return tool_tip_items, formatters


def _get_color_palette(series_count):
    palette_size = min(256, series_count + int(series_count / 5))
    return viridis(palette_size), palette_size


def _wrap_df_columns(data: pd.DataFrame, wrap_len: int = 50):
    """Wrap any string columns."""
    if not data.empty:
        for col in data.columns:
            if isinstance(data[col].iloc[0], str):
                data[col] = data[col].str.wrap(wrap_len)


def _get_tick_formatter() -> DatetimeTickFormatter:
    """Return tick formatting for different zoom levels."""
    # '%H:%M:%S.%3Nms
    tick_format = DatetimeTickFormatter()
    tick_format.days = ["%m-%d %H:%M"]
    tick_format.hours = ["%H:%M:%S"]
    tick_format.minutes = ["%H:%M:%S"]
    tick_format.seconds = ["%H:%M:%S"]
    tick_format.milliseconds = ["%H:%M:%S.%3N"]
    return tick_format


def _calc_auto_plot_height(group_count):
    """Dynamic calculation of plot height."""
    ht_per_row = 25 if group_count > 15 else 40
    return max(ht_per_row * group_count, 300)


# pylint: disable=too-many-arguments, invalid-name, too-many-locals
def _create_range_tool(
    data,
    min_time,
    max_time,
    plot_range,
    width,
    height,
    time_column: str = None,
    y: str = "y_index",
):
    """Create plot bar to act as as range selector."""
    ext_min = min_time - ((max_time - min_time) * 0.15)
    ext_max = max_time + ((max_time - min_time) * 0.15)
    plot_height = max(120, int(height * 0.20))
    rng_select = figure(
        x_range=(ext_min, ext_max),
        title="Range Selector",
        plot_height=plot_height,
        plot_width=width,
        x_axis_type="datetime",
        y_axis_type=None,
        tools="",
        toolbar_location=None,
    )
    help_str = (
        "Drag the middle or edges of the selection box to change "
        + "the range in the main chart"
    )
    rng_select.add_layout(
        Title(text=help_str, align="right", text_font_size="10px"), "below"
    )
    rng_select.xaxis[0].formatter = _get_tick_formatter()
    if isinstance(data, dict):
        for _, series_def in data.items():
            rng_select.circle(
                x=series_def["time_column"],
                y=y,
                color=series_def["color"],
                source=series_def["source"],
            )
    elif isinstance(data, pd.DataFrame):
        rng_select.circle(
            x=time_column, y=y, color="blue", source=ColumnDataSource(data)
        )

    range_tool = RangeTool(x_range=plot_range)
    range_tool.overlay.fill_color = "navy"
    range_tool.overlay.fill_alpha = 0.2
    rng_select.ygrid.grid_line_color = None
    rng_select.add_tools(range_tool)
    rng_select.toolbar.active_multi = range_tool
    return rng_select


# pylint: enable=too-many-arguments


def _add_ref_line(plot, ref_time, ref_text="Ref time", series_count=1, index=0):
    """Add a reference marker line and label at `ref_time`."""
    ref_label_tm = pd.Timestamp(ref_time)
    index = min(index, series_count)
    plot.line(
        x=[ref_label_tm, ref_label_tm],
        y=[0, series_count],
        line_width=1,
        line_color="red",
        line_dash="dashed",
        line_alpha=0.5,
    )
    ref_label = Label(
        x=ref_label_tm,
        y=0,
        y_offset=10 + (10 * index),
        x_units="data",
        y_units="data",
        text=f"< {ref_text}",
        text_font_size="8pt",
        text_alpha=0.5,
        render_mode="css",
        border_line_color="red",
        border_line_alpha=0.3,
        background_fill_color="white",
        background_fill_alpha=0.3,
    )

    plot.add_layout(ref_label)


def _plot_ref_events(
    plot: figure,
    time_col: str,
    group_count: int,
    ref_events: Optional[pd.DataFrame] = None,
    ref_col: Optional[str] = None,
    ref_times: Optional[List[Tuple[datetime, str]]] = None,
):
    """Plot reference lines/labels."""
    if ref_events is not None:
        if isinstance(ref_events, pd.Series):
            ref_events = pd.DataFrame(ref_events)
        for idx, event in enumerate(ref_events.itertuples()):
            evt_time = event._asdict()[time_col]
            evt_label = (
                event._asdict()[ref_col] if ref_col else f"reference {event.Index}"
            )
            _add_ref_line(
                plot=plot,
                ref_time=evt_time,
                ref_text=evt_label,
                series_count=group_count,
                index=idx,
            )
    elif ref_times:
        for idx, (evt_time, evt_label) in enumerate(ref_times):
            evt_label = evt_label or f"reference {idx}"
            _add_ref_line(
                plot=plot,
                ref_time=evt_time,
                ref_text=evt_label,
                series_count=group_count,
                index=idx,
            )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Creates an entity graph for a Microsoft Sentinel Incident."""
from datetime import datetime
from typing import List, Optional, Union

import numpy as np
import networkx as nx
import pandas as pd
from bokeh.io import output_notebook, show
from bokeh.layouts import column
from bokeh.models import LayoutDOM
from bokeh.plotting import figure, from_networkx
from bokeh.models import Circle, HoverTool, Label

from .._version import VERSION
from ..common.exceptions import MsticpyUserError
from ..datamodel.entities import Entity
from ..datamodel.entities.alert import Alert
from ..datamodel.soc.incident import Incident
from ..nbtools.security_alert import SecurityAlert
from ..nbtools.timeline import display_timeline
from ..nbtools.timeline_duration import display_timeline_duration

__version__ = VERSION
__author__ = "Pete Bryan"

req_alert_cols = ["DisplayName", "Severity", "AlertType"]
req_inc_cols = ["id", "name", "properties.severity"]


class EntityGraph:
    """Create a graph for visualizing and tracking links between entities."""

    def __init__(
        self,
        entity: Union[Incident, Alert, pd.DataFrame, pd.Series, Entity, SecurityAlert],
    ):
        """
        Create a new instance of the entity graph.

        Parameters
        ----------
        entity : Union[Incident, Alert, pd.DataFrame, pd.Series, Entity, SecurityAlert]
            The initial item to add to the graph.
            Can be an Incident, Alert, SecurityAlert or other Entity

        """
        output_notebook()
        self.alertentity_graph = nx.Graph(id="IncidentGraph")
        if isinstance(entity, (Incident, Alert)):
            self._add_incident_or_alert_node(entity)
        elif isinstance(entity, pd.DataFrame):
            self.add_incident(entity)
        elif isinstance(entity, pd.Series):
            self.add_incident(entity.to_frame().T)
        elif isinstance(entity, Entity):
            self._add_entity_node(entity)
        elif isinstance(entity, SecurityAlert):
            entity = Alert(entity)  # type: ignore
            self._add_incident_or_alert_node(entity)

    def plot(self, hide: bool = False, timeline: bool = False, **kwargs) -> LayoutDOM:
        """
        Plot a graph of entities.

        Parameters
        ----------
        hide : bool, optional
            Set true to not display the graphic, by default False
        timeline : bool, optional
            Set to True to display a timeline, by default False
        node_size : int, optional
            Size of the nodes in pixels, by default 25
        font_size : int, optional
            Font size for node labels, by default 10
            Can be an integer (point size) or a string (e.g. "10pt")
        width : int, optional
            Width in pixels, by default 800
        height : int, optional
            Image height (the default is 800)
        scale : int, optional
            Position scale (the default is 2)

        Returns
        -------
        LayoutDOM
            A Bokeh figure object

        """
        if timeline:
            return self._plot_with_timeline(hide=hide, **kwargs)
        return self._plot_no_timeline(hide=hide, **kwargs)

    def _plot_no_timeline(self, hide: bool = False, **kwargs) -> LayoutDOM:
        """
        Plot a graph of entities.

        Parameters
        ----------
        hide : bool, optional
            Set true to not display the graphic, by default False

        Returns
        -------
        LayoutDOM
            A Bokeh figure object

        """
        return plot_entitygraph(self.alertentity_graph, hide=hide, **kwargs)

    def _plot_with_timeline(self, hide: bool = False, **kwargs) -> LayoutDOM:
        """
        Plot the entity graph with a timeline.

        Parameters
        ----------
        hide : bool, optional
            Set true to not display the graphic, by default False

        Returns
        -------
        LayoutDOM
            A Bokeh figure object

        """
        timeline = None
        tl_df = self.to_df()
        tl_type = "duration"
        if len(tl_df["EndTime"].unique()) == 1 and not tl_df["EndTime"].unique()[0]:
            tl_type = "discreet"
            if (
                len(tl_df["TimeGenerated"].unique()) == 1
                and not tl_df["TimeGenerated"].unique()[0]
            ):
                print("No timestamps available to create timeline")
                return self._plot_no_timeline(timeline=False, hide=hide, **kwargs)
        tl_df["TimeGenerated"] = pd.to_datetime(tl_df["TimeGenerated"], utc=True)
        tl_df["StartTime"] = pd.to_datetime(tl_df["StartTime"], utc=True)
        tl_df["EndTime"] = pd.to_datetime(tl_df["EndTime"], utc=True)
        graph = self._plot_no_timeline(hide=True, **kwargs)
        if tl_type == "duration":
            timeline = display_timeline_duration(
                tl_df.dropna(subset=["TimeGenerated"]),
                group_by="Name",
                title="Entity Timeline",
                time_column="StartTime",
                end_time_column="EndTime",
                source_columns=["Name", "Description", "Type", "TimeGenerated"],
                hide=True,
                width=800,
            )
        elif tl_type == "discreet":
            timeline = display_timeline(
                tl_df.dropna(subset=["TimeGenerated"]),
                group_by="Type",
                title="Entity Timeline",
                time_column="TimeGenerated",
                source_columns=["Name", "Description", "Type", "TimeGenerated"],
                hide=True,
                width=800,
            )
        plot_layout = column(graph, timeline) if timeline else graph
        if not hide:
            show(plot_layout)
        return plot_layout

    def add_entity(self, ent: Entity, attached_to: str = None):
        """
        Add an entity to the graph.

        Parameters
        ----------
        ent : Entity
            The entity object to add the graph
        attached_to : str, optional
            The name of the node to attach the entity to, by default None

        """
        self._add_entity_node(ent, attached_to)

    def add_incident(self, incident: Union[Incident, Alert, pd.DataFrame]):
        """
        Add another incident or set of incidents to the graph.

        Parameters
        ----------
        incident : Union[Incident, Alert, pd.DataFrame]
            This can be an alert, and incident or a DataFrame of alerts or incidents

        """
        inc = None
        if isinstance(incident, pd.DataFrame):
            for row in incident.iterrows():
                if "name" in row[1]:
                    inc = Incident(src_event=row[1])
                elif "AlertName" in row[1]:
                    inc = Alert(src_event=row[1])  # type: ignore
                self._add_incident_or_alert_node(inc)
        else:
            self._add_incident_or_alert_node(incident)

    def add_note(
        self,
        name: str,
        description: Optional[str] = None,
        attached_to: Union[str, List] = None,
    ):
        """
        Add a node to the graph representing a note or comment.

        Parameters
        ----------
        name : str
            The name of the node to add
        description : Optional[str], optional
            A description of the note, by default None
        attached_to : Union[str, List], optional
            What existing nodes on the graph to attach it the note to, by default None
        user: str, optional
            What user to associate the note with

        """
        self.alertentity_graph.add_node(
            name,
            Name=name,
            Description=description,
            Type="analystnote",
            TimeGenerated=datetime.now(),
        )
        if attached_to:
            if isinstance(attached_to, str):
                attached_to = [attached_to]
            for link in attached_to:
                self.add_link(name, link)

    def add_link(self, source: str, target: str):
        """
        Add a link between 2 nodes on the graph.

        Parameters
        ----------
        source : str
            Name of node to link from
        target : str
            Name of node to link to

        Raises
        ------
        MsticpyUserError
            If nodes aren't present in the graph


        """
        # Check names are present
        if (
            source in self.alertentity_graph.nodes()
            and target in self.alertentity_graph.nodes()
        ):
            self.alertentity_graph.add_edge(source, target)
        else:
            missing = [
                name
                for name in [source, target]
                if name not in self.alertentity_graph.nodes()
            ]
            raise MsticpyUserError(title=f"Node(s) {missing} not found in graph")

    def remove_link(self, source: str, target: str):
        """
        Remove a link between 2 nodes on the graph.

        Parameters
        ----------
        source : str
            Name of node to remove link from
        target : str
            name of node to remove link to

        Raises
        ------
        MsticpyUserError
            If edge isn't present in the graph

        """
        if (
            source in self.alertentity_graph.nodes()
            and target in self.alertentity_graph.nodes()
            and self.alertentity_graph.has_edge(source, target)
        ):
            self.alertentity_graph.remove_edge(source, target)
        else:
            raise MsticpyUserError(
                title=f"No edge exists between {source} and {target}"
            )

    def remove_node(self, name: str):
        """
        Remove a node from the graph.

        Parameters
        ----------
        name : str
            The name of the node to remove.

        """
        # Check node is present
        if name in self.alertentity_graph.nodes():
            self.alertentity_graph.remove_node(name)
        else:
            raise MsticpyUserError(f"Node named {name} not found")

    def to_df(self) -> pd.DataFrame:
        """Generate a dataframe of nodes in the graph."""
        names = [node[1]["Name"] for node in self.alertentity_graph.nodes.items()]
        descs = [
            node[1]["Description"] for node in self.alertentity_graph.nodes.items()
        ]
        types = [node[1]["Type"] for node in self.alertentity_graph.nodes.items()]
        times = [
            node[1]["TimeGenerated"] if "TimeGenerated" in node[1] else None
            for node in self.alertentity_graph.nodes.items()
        ]
        starttimes = [
            node[1]["StartTime"] if "StartTime" in node[1] else node[1]["TimeGenerated"]
            for node in self.alertentity_graph.nodes.items()
        ]
        endtimes = [
            node[1]["EndTime"] if "EndTime" in node[1] else None
            for node in self.alertentity_graph.nodes.items()
        ]
        tl_df = pd.DataFrame(
            {
                "Name": names,
                "Description": descs,
                "Type": types,
                "TimeGenerated": times,
                "EndTime": endtimes,
                "StartTime": starttimes,
            }
        )
        tl_df.replace("None", np.NaN, inplace=True)
        return tl_df

    def _add_incident_or_alert_node(self, incident: Union[Incident, Alert, None]):
        """Check what type of entity is passed in and creates relevent graph."""
        if isinstance(incident, Incident):
            self._add_incident_node(incident)
        elif isinstance(incident, Alert):
            self._add_alert_node(incident)

    def _add_entity_node(self, ent, attached_to=None):
        """Add an Entity to the graph."""
        self.alertentity_graph = nx.compose(self.alertentity_graph, ent.to_networkx())
        if attached_to:
            self.add_link(attached_to, ent.name_str)

    def _add_alert_node(self, alert, incident_name=None):
        """Add an alert entity to the graph."""
        self.alertentity_graph = nx.compose(self.alertentity_graph, alert.to_networkx())
        if alert["Entities"]:
            for ent in alert["Entities"]:
                self._add_entity_node(ent, alert.name_str)
        if incident_name:
            self.add_link(incident_name, alert.name_str)

    def _add_incident_node(self, incident):
        """Add an incident entity to the graph."""
        self.alertentity_graph = nx.compose(
            self.alertentity_graph, incident.to_networkx()
        )
        if incident.Alerts:
            for alert in incident.Alerts:
                self._add_alert_node(alert, incident.name_str)
        if incident.Entities:
            entities = _dedupe_entities(incident.Alerts, incident.Entities)
            for ent in entities:
                self._add_entity_node(ent, incident.name_str)

    def _add_entity_edges(self, edges: set, attached_to: str):
        """Check entity edges and add them."""
        for edge in edges:
            if isinstance(edge.target, Entity):
                if not self.alertentity_graph.has_node(edge.target.name_str):
                    self._add_entity_node(edge.target)
                try:
                    self.add_link(attached_to, edge.target.name_str)
                except MsticpyUserError:
                    pass

    @property
    def graph(self) -> nx.Graph:
        """Return the raw NetworkX graph."""
        return self.alertentity_graph


def _dedupe_entities(alerts, ents) -> list:
    """Deduplicate incident and alert entities."""
    alrt_ents = []
    for alrt in alerts:
        if alrt["Entities"]:

            alrt_ents += [ent.__hash__() for ent in alrt["Entities"]]
    for ent in ents:
        if ent.__hash__() in alrt_ents:
            ents.remove(ent)
    return ents


def plot_entitygraph(  # pylint: disable=too-many-locals
    entity_graph: nx.Graph,
    node_size: int = 25,
    font_size: Union[int, str] = 10,
    height: int = 800,
    width: int = 800,
    scale: int = 2,
    hide: bool = False,
) -> figure:
    """
    Plot entity graph with Bokeh.

    Parameters
    ----------
    entity_graph : nx.Graph
        The entity graph as a networkX graph
    node_size : int, optional
        Size of the nodes in pixels, by default 25
    font_size : int, optional
        Font size for node labels, by default 10
        Can be an integer (point size) or a string (e.g. "10pt")
    width : int, optional
        Width in pixels, by default 800
    height : int, optional
        Image height (the default is 800)
    scale : int, optional
        Position scale (the default is 2)
    hide : bool, optional
        Don't show the plot, by default False. If True, just
        return the figure.

    Returns
    -------
    bokeh.plotting.figure
        The network plot.

    """
    color_map = {
        "incident": "red",
        "alert": "orange",
        "alerts": "orange",
        "securityalert": "orange",
        "analystnote": "blue",
    }
    output_notebook()
    font_pnt = f"{font_size}pt" if isinstance(font_size, int) else font_size
    node_attrs = {}
    for node, attrs in entity_graph.nodes(data=True):
        try:
            color = color_map.get(attrs["Type"].lower(), "green")
        except KeyError:
            color = "green"
        node_attrs.update({node: color})

    nx.set_node_attributes(entity_graph, node_attrs, "node_color")

    plot = figure(
        title="Alert Entity graph",
        x_range=(-3, 3),
        y_range=(-3, 3),
        width=width,
        height=height,
    )

    plot.add_tools(
        HoverTool(
            tooltips=[
                ("Name", "@Name"),
                ("Description", "@Description"),
                ("Type", "@Type"),
            ]
        )
    )

    graph_renderer = from_networkx(
        entity_graph, nx.spring_layout, scale=scale, center=(0, 0)
    )

    graph_renderer.node_renderer.glyph = Circle(
        size=node_size, fill_color="node_color", fill_alpha=0.5
    )
    # pylint: disable=no-member
    plot.renderers.append(graph_renderer)

    # Create labels
    for name, pos in graph_renderer.layout_provider.graph_layout.items():
        label = Label(
            x=pos[0],
            y=pos[1],
            x_offset=5,
            y_offset=5,
            text=name,
            text_font_size=font_pnt,
        )
        plot.add_layout(label)
    # pylint: enable=no-member
    if not hide:
        show(plot)
    return plot



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module docstring."""
from typing import Iterable, Optional, Tuple, Union

import pandas as pd
from bokeh.models import LayoutDOM
from bokeh.plotting import figure

from .._version import VERSION
from ..common.exceptions import MsticpyUserError
from ..nbtools.process_tree import build_and_show_process_tree
from ..nbtools.timeline import display_timeline, display_timeline_values
from ..nbtools.timeline_duration import display_timeline_duration
from .entity_graph_tools import EntityGraph, req_alert_cols, req_inc_cols
from .matrix_plot import plot_matrix

__version__ = VERSION
__author__ = "Ian Hellen"


@pd.api.extensions.register_dataframe_accessor("mp_plot")
class MsticpyPlotAccessor:
    """Pandas api extension for MSTICPy visualizations."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        self._df = pandas_obj

    def timeline(self, **kwargs) -> LayoutDOM:
        """
        Display a timeline of events.

        Parameters
        ----------
        time_column : str, optional
            Name of the timestamp column
            (the default is 'TimeGenerated')
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)

        Other Parameters
        ----------------
        title : str, optional
            Title to display (the default is None)
        alert : SecurityAlert, optional
            Add a reference line/label using the alert time (the default is None)
        ref_event : Any, optional
            Add a reference line/label using the alert time (the default is None)
        ref_time : datetime, optional
            Add a reference line/label using `ref_time` (the default is None)
        group_by : str
            The column to group timelines on.
        legend: str, optional
            "left", "right", "inline" or "none"
            (the default is to show a legend when plotting multiple series
            and not to show one when plotting a single series)
        yaxis : bool, optional
            Whether to show the yaxis and labels (default is False)
        ygrid : bool, optional
            Whether to show the yaxis grid (default is False)
        xgrid : bool, optional
            Whether to show the xaxis grid (default is True)
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy")
        overlay_data : pd.DataFrame:
            A second dataframe to plot as a different series.
        overlay_color : str
            Overlay series color (default is "green")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_time_col : str, optional
            Add references line/label using the this column in `ref_events`
            for the time value (x-axis).
            (this defaults the value of the `time_column` parameter or 'TimeGenerated'
            `time_column` is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return display_timeline(data=self._df, **kwargs)

    # pylint: disable=invalid-name
    def timeline_values(self, value_col: str = None, **kwargs) -> LayoutDOM:
        """
        Display a timeline of events.

        Parameters
        ----------
        time_column : str, optional
            Name of the timestamp column
            (the default is 'TimeGenerated')
        value_col : str
            The column name holding the value to plot vertically
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)

        Other Parameters
        ----------------
        x : str, optional
            alias of `time_column`
        y : str, optional
            alias of `value_col`
        title : str, optional
            Title to display (the default is None)
        ref_event : Any, optional
            Add a reference line/label using the alert time (the default is None)
        ref_time : datetime, optional
            Add a reference line/label using `ref_time` (the default is None)
        ref_label : str, optional
            A label for the `ref_event` or `ref_time` reference item
        group_by : str
            (where `data` is a DataFrame)
            The column to group timelines on
        legend: str, optional
            "left", "right", "inline" or "none"
            (the default is to show a legend when plotting multiple series
            and not to show one when plotting a single series)
        yaxis : bool, optional
            Whether to show the yaxis and labels
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy"). This is overridden by
            automatic color assignments if plotting a grouped chart
        kind : Union[str, List[str]]
            one or more glyph types to plot., optional
            Supported types are "circle", "line" and "vbar" (default is "vbar")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_time_col : str, optional
            Add references line/label using the this column in `ref_events`
            for the time value (x-axis).
            (this defaults the value of the `time_column` parameter or 'TimeGenerated'
            `time_column` is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return display_timeline_values(data=self._df, value_col=value_col, **kwargs)

    def timeline_duration(
        self,
        group_by: Union[Iterable[str], str],
        time_column: str = "TimeGenerated",
        end_time_column: Optional[str] = None,
        **kwargs,
    ) -> LayoutDOM:  # noqa: C901, MC0001
        """
        Display a duration timeline of events grouped by one or more columns.

        Parameters
        ----------
        group_by : Union[Iterable[str], str]
            The column name or iterable of column names to group the data by.
        time_column : str
            Primary time column - will be used to calculate the
            start time of the duration for each group.
            If `end_time_column` is not specified it will also be used to
            calculate the end time.
        end_time_column : Optional[str]
            If supplied, it will be used to calculate the end time
            of the duration for each group.

        Other Parameters
        ----------------
        title : str, optional
            Title to display (the default is None)
        ylabel_cols : Optional[Iterable[str]], optional
            The subset of the group columns to use for the y-axis labels.
        yaxis : bool, optional
            Whether to show the yaxis and labels
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return display_timeline_duration(
            data=self._df,
            group_by=group_by,
            time_column=time_column,
            end_time_column=end_time_column,
            **kwargs,
        )

    def process_tree(self, **kwargs) -> Tuple[figure, LayoutDOM]:
        """
        Build and plot a process tree.

        Parameters
        ----------
        schema : ProcSchema, optional
            The data schema to use for the data set, by default None
            (if None the schema is inferred)
        output_var : str, optional
            Output variable for selected items in the tree,
            by default None
        legend_col : str, optional
            The column used to color the tree items, by default None
        show_table: bool
            Set to True to show a data table, by default False.

        Other Parameters
        ----------------
        height : int, optional
            The height of the plot figure
            (the default is 700)
        width : int, optional
            The width of the plot figure (the default is 900)
        title : str, optional
            Title to display (the default is None)
        hide_legend : bool, optional
            Hide the legend box, even if legend_col is specified.
        pid_fmt : str, optional
            Display Process ID as 'dec' (decimal) or 'hex' (hexadecimal),
            default is 'hex'.

        Returns
        -------
        Tuple[figure, LayoutDOM]:
            figure - The main bokeh.plotting.figure
            Layout - Bokeh layout structure.

        """
        return build_and_show_process_tree(data=self._df, **kwargs)

    def matrix(self, **kwargs) -> LayoutDOM:
        """
        Plot data as an intersection matrix.

        Parameters
        ----------
        data : pd.DataFrame
            The data to plot.
        x : str
            Column to plot on the x (horizontal) axis
        x_col : str
            Alias for 'x'
        y : str
            Column to plot on the y (vertical) axis
        y_col : str
            Alias for 'y'
        title : str, optional
            Custom title, default is 'Intersection plot'
        value_col : str, optional
            Column from the DataFrame used to size the intersection points.
        dist_count : bool, optional
            Calculates a count of distinct values (from `value_col`) and uses
            this to size the intersection points.
            Requires `value_col` to be specified.
        log_size : bool, optional
            Takes the log of the size value before calculating the intersection
            display point size.
            Can be combined with `invert`.
        invert : bool, optional
            Takes the inverse of the size value as the basis for calculating
            the intersection display point size. This is useful for highlighting
            rare interactions.
            Can be combined with `log_size`.
        intersect : bool, optional
            Plots points of a fixed size, rather than using a sizing value. This
            is useful for just showing the presence/absence of an interaction.
        height : int, optional
            The plot height. Default is 700
        width : int
            The plot width. Default is 900
        color : str
            The color of the plotted points, default is "red"
        sort : Union[str, bool], optional
            Sorts the labels of both axes, default is None.
            Acceptable values are:
            'asc' (or string starting with 'asc') - Sort ascending
            'desc' (or string starting with 'asc') - Sort descending
            False or None (no sort)
            True  - Sort ascending
        sort_x : str, optional
            Sorts the labels of the x axis (takes precedence over `sort`),
            default is None.
            Acceptable values are:
            'asc' (or string starting with 'asc') - Sort ascending
            'desc' (or string starting with 'asc') - Sort descending
            False or None (no sort)
            True  - Sort ascending
        sort_y : str, optional
            Sorts the labels of the y axis (takes precedence over `sort`),
            default is None.
            Acceptable values are:
            'asc' (or string starting with 'asc') - Sort ascending
            'desc' (or string starting with 'asc') - Sort descending
            False or None (no sort)
            True  - Sort ascending
        hide : bool, optional
            Creates and returns but does not display the plot, default
            is False.
        font_size : int, optional
            Manually specify the font size for axis labels, the
            default is to automatically calculate a size based on the
            number of items in each axis.
        max_label_font_size : int, optional
            The maximum size, in points, of the X and Y labels, default is 11.


        Returns
        -------
        LayoutDOM
            The Bokeh plot

        """
        return plot_matrix(data=self._df, **kwargs)

    def incident_graph(
        self, timeline: bool = False, hide: bool = False, **kwargs
    ) -> LayoutDOM:
        """
        Plot an incident graph if the dataframe contains incidents or alerts.

        Parameters
        ----------
        timeline : bool, optional
            True to plot the entity timeline, by default False
        hide : bool, optional
            True to hide the plot, by default False

        Other Parameters
        ----------------
        node_size : int, optional
            Size of the nodes in pixels, by default 25
        font_size : int, optional
            Font size for node labels, by default 10
            Can be an integer (point size) or a string (e.g. "10pt")
        width : int, optional
            Width in pixels, by default 800
        height : int, optional
            Image height (the default is 800)
        scale : int, optional
            Position scale (the default is 2)

        Raises
        ------
        MsticpyUserError
            Raised if the dataframe does not contain incidents or alerts.

        """
        if not all(elem in self._df.columns for elem in req_alert_cols) and any(
            elem not in self._df.columns for elem in req_inc_cols
        ):
            raise MsticpyUserError("DataFrame must consist of Incidents or Alerts")
        graph = EntityGraph(self._df)
        return graph.plot(hide=hide, timeline=timeline, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for Log Analytics-related configuration."""

import os
import json
from typing import Dict, Any, Optional
from pathlib import Path

from IPython.display import display
import ipywidgets as widgets

from .exceptions import MsticpyUserConfigError
from .utility import export, is_valid_uuid, md, md_warn
from . import pkg_config
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


_RESOURCES = [
    (
        "https://github.com/Azure/Azure-Sentinel-Notebooks/blob/"
        "master/ConfiguringNotebookEnvironment.ipynb"
    ),
    (
        "https://github.com/Azure/Azure-Sentinel-Notebooks/blob/"
        "master/A%20Getting%20Started%20Guide%20For%20Azure%20"
        "Sentinel%20ML%20Notebooks.ipynb"
    ),
]

_NO_CONFIG_WARN = [
    "Could not find Microsoft Sentinel settings in msticpyconfig.yaml or a config.json ",
    "(in the current directory or via a MSTICPYCONFIG variable.)",
    "We found the file '{config_file}' and will use this.",
    "We recommend using an explicit msticpyconfig.yaml specified using the",
    "MSTICPYCONFIG environment variable. See:",
    (
        "https://msticpy.readthedocs.io/en/latest/data_acquisition/DataProviders.html#"
        "connecting-to-an-azure-sentinel-workspace"
    ),
    "and",
    "https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html",
    "for more details.",
    "Also see the notebooks:",
    *_RESOURCES,
]

_NO_CONFIG_ERR = [
    "Could not find msticpyconfig.yaml or config.json.",
    "The 'config.json' file is created when you launch notebooks from "
    "Microsoft Sentinel. If you have copied the notebook to another location "
    "or folder you will need to copy this configuration file.",
    "Alternatively, we recommend using an explicit msticpyconfig.yaml"
    "and adding your Workspace and Tenant IDs to that file.",
    "",
    "You can create a settings file using the following commands:",
    ">>> from msticpy.config import MpConfigEdit",
    ">>> MpConfigEdit()",
]

WIDGET_DEFAULTS = {
    "layout": widgets.Layout(width="95%"),
    "style": {"description_width": "150px"},
}


@export
class WorkspaceConfig:
    """Workspace configuration class."""

    # Constants
    TENANT_ID = "{{cookiecutter.tenant_id}}"
    SUBSCRIPTION_ID = "{{cookiecutter.subscription_id}}"
    RESOURCE_GROUP = "{{cookiecutter.resource_group}}"
    WORKSPACE_ID = "{{cookiecutter.workspace_id}}"
    WORKSPACE_NAME = "{{cookiecutter.workspace_name}}"

    PKG_CONF_TENANT_KEY = "TenantId"
    PKG_CONF_WS_KEY = "WorkspaceId"

    CONF_WS_ID_KEY = "workspace_id"
    CONF_TENANT_ID_KEY = "tenant_id"
    CONF_SUB_ID_KEY = "subscription_id"
    CONF_RES_GROUP_KEY = "resource_group"
    CONF_WS_NAME_KEY = "workspace_name"

    def __init__(
        self,
        workspace: Optional[str] = None,
        config_file: Optional[str] = None,
        interactive: bool = True,
    ):
        """
        Load current Azure Notebooks configuration for Log Analytics.

        Parameters
        ----------
        config_file : Optional[str], optional
            path to a configuration file,
            If not specified, the defaults is to use a configured msticpyconfig.yaml
            If this isn't configured, it will search for (first) a config.json
            and (second) a msticpyconfig.yaml in (first) the current directory
            and (second) the parent directory and subfolders.
        workspace : str, Optional[str]
            Workspace name (where multiple workspaces are configured),
            by default the Default workspace will be used.
        interactive : bool, optional
            If this is False, initializing the class will not raise an
            exception if no configuration is found. By default, True.

        """
        self._config: Dict[str, str] = {}
        self._interactive = interactive
        self._config_file = config_file
        # If config file specified, use that
        if config_file:
            self._config.update(self._read_config_values(config_file))
        else:
            self._determine_config_source(workspace)

    def _determine_config_source(self, workspace):
        # First, try default MSTICPy config
        self._read_pkg_config_values(workspace_name=workspace)
        if self.config_loaded:
            return
        # Next, search for a config.json in the current director
        if Path("./config.json").exists():
            self._config_file = "./config.json"
        else:
            self._config_file = self._search_for_file("**/config.json")
        if self._config_file:
            self._config.update(self._read_config_values(self._config_file))
            return

        # Finally, search for a msticpyconfig.yaml
        if (
            os.environ.get("MSTICPYCONFIG")
            and Path(os.environ.get("MSTICPYCONFIG")).exists()
        ):
            self._config_file = os.environ.get("MSTICPYCONFIG")
        elif Path("./msticpyconfig.yaml").exists():
            self._config_file = "./msticpyconfig.yaml"
        else:
            self._config_file = self._search_for_file("**/msticpyconfig.yaml")
        if self._config_file:
            os.environ["MSTICPYCONFIG"] = self._config_file
            pkg_config.refresh_config()
            self._read_pkg_config_values(workspace_name=workspace)
            return
        # Finally, throw an exception (unless non-interactive)
        if self._interactive:
            # If we've arrived here after searching current folder and parent
            # then we give up. (We create but don't raise an actual exception)
            display(
                MsticpyUserConfigError(
                    *_NO_CONFIG_ERR,
                    title="Workspace configuration missing.",
                    **{f"nb_{idx}_uri": res for idx, res in enumerate(_RESOURCES)},
                )
            )

    def __getitem__(self, key: str):
        """Allow property get using dictionary key syntax."""
        if key in self._config:
            return self._config[key]
        raise KeyError

    def __setitem__(self, key: str, value: Any):
        """Allow property set using dictionary key syntax."""
        self._config[key] = value

    def __contains__(self, key: str):
        """Allow property in test."""
        # In operator overload
        return key == "Type" or key in self._config or key in self.__dict__

    def __repr__(self):
        """Return contents of current config."""
        return self._config.__repr__()

    @property
    def config_loaded(self) -> bool:
        """
        Return True if workspace id and tenant id have values.

        Returns
        -------
        bool
            True if configuration loaded.

        """
        ws_value = self._config.get(self.CONF_WS_ID_KEY, None)
        ten_value = self._config.get(self.CONF_TENANT_ID_KEY, None)
        return is_valid_uuid(ws_value) and is_valid_uuid(ten_value)  # type: ignore

    @property
    def code_connect_str(self) -> str:
        """
        Return the Log Analytics connection string for dev code auth.

        Returns
        -------
        str
            Connection string

        """
        ten_id = self._config.get(self.CONF_TENANT_ID_KEY, None)
        ws_id = self._config.get(self.CONF_WS_ID_KEY, None)
        if not ten_id:
            raise KeyError(
                f"Configuration setting for {self.CONF_TENANT_ID_KEY} "
                + "could not be found."
            )
        if not ws_id:
            raise KeyError(
                f"Configuration setting for {self.CONF_WS_ID_KEY} "
                + "could not be found."
            )
        return f"loganalytics://code().tenant('{ten_id}').workspace('{ws_id}')"

    @staticmethod
    def _read_config_values(file_path: str) -> Dict[str, str]:
        """Read configuration file."""
        if not file_path:
            return {}
        try:
            with open(file_path, "r", encoding="utf-8") as json_file:
                if json_file:
                    return json.load(json_file)
        except json.JSONDecodeError:
            pass
        return {}

    @classmethod
    def list_workspaces(cls) -> Dict:
        """
        Return list of available workspaces.

        Returns
        -------
        Dict
            Dictionary of workspaces with workspace and tenantIds.

        """
        ws_settings = pkg_config.settings.get("AzureSentinel", {}).get("Workspaces")
        if not ws_settings:
            return {}
        return {
            ws_name: {
                cls.PKG_CONF_WS_KEY: ws.get(cls.PKG_CONF_WS_KEY),
                cls.PKG_CONF_TENANT_KEY: ws.get(cls.PKG_CONF_TENANT_KEY),
            }
            for ws_name, ws in ws_settings.items()
        }

    def prompt_for_ws(self):
        """Display an interactive prompt for Workspace details."""
        md_warn("No Microsoft Sentinel configuration found.")
        md(
            "Please enter the workspace ID and tenant Id"
            + " to allow connection to the Microsoft Sentinel workspace."
        )
        ws_id_wgt = widgets.Text(description="Workspace Id:", **WIDGET_DEFAULTS)
        ten_id_wgt = widgets.Text(description="Tenant Id:", **WIDGET_DEFAULTS)

        def update_ws(chg):
            self["workspace_id"] = chg.get("new")

        def update_tnt(chg):
            self["tenant_id"] = chg.get("new")

        ws_id_wgt.observe(update_ws, names="value")
        ten_id_wgt.observe(update_tnt, names="value")
        display(widgets.VBox([ws_id_wgt, ten_id_wgt]))
        md(
            (
                "You can avoid this prompt in future by following the"
                " guidance in the"
                " <a href='https://msticpy.readthedocs.io/en/latest/"
                "data_acquisition/DataProviders.html#"
                "connecting-to-an-azure-sentinel-workspace' "
                "target='_blank' rel='noopener noreferrer'>"
                "Connecting to a Microsoft Sentinel Workspace</a>"
            ),
        )

    def _read_pkg_config_values(self, workspace_name: str = None):
        as_settings = pkg_config.settings.get("AzureSentinel")
        if not as_settings:
            return {}
        ws_settings = as_settings.get("Workspaces")  # type: ignore
        if not ws_settings:
            return {}
        if workspace_name and workspace_name in ws_settings:
            selected_workspace = ws_settings[workspace_name]
        elif "Default" in ws_settings:
            selected_workspace = ws_settings["Default"]
        else:
            return {}
        if (
            selected_workspace
            and self.PKG_CONF_WS_KEY in selected_workspace
            and self.PKG_CONF_TENANT_KEY in selected_workspace
        ):
            self._config[self.CONF_WS_ID_KEY] = selected_workspace.get(
                self.PKG_CONF_WS_KEY
            )
            self._config[self.CONF_TENANT_ID_KEY] = selected_workspace.get(
                self.PKG_CONF_TENANT_KEY
            )
        return {}

    def _search_for_file(self, pattern: str) -> Optional[str]:
        config_file = None
        for start_path in (".", ".."):
            searched_configs = list(Path(start_path).glob(pattern))
            for found_file in searched_configs:
                if found_file.name == "msticpyconfig.yaml":
                    config_file = str(found_file)
                    break
                test_content = self._read_config_values(str(found_file))
                if "workspace_id" in test_content:
                    config_file = str(found_file)
                    break
            if config_file:
                break
        if config_file:
            # Warn that we're using a "found" file, not one in the current directory
            print("Warning")
            print("\n".join(_NO_CONFIG_WARN).format(config_file=config_file))
        return config_file



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
base64_unpack.

The main function of this module is to decode and unpack
strings that are obfuscated using base64 and/or certain
compression algorithms such as gzip and zip.

It has the following functions:
unpack_items - this is the main entry point and takes either a string or
a pandas dataframe (with specified column) as input. It returns a string
with obfuscated parts replaced by decoded equivalents (unless the decoding
results in an undecodable binary, in which case a placeholder is used).


Other helper functions may also be useful standalone
get_items_from_gzip(binary): Return decompressed gzip content of byte string
get_items_from_zip(binary): Return dictionary of zip contents from byte string
get_items_from_tar(binary): Return dictionary of tar file contents
get_hashes(binary): Return md5, sha1 and sha256 hashes of input byte string
"""

import base64
import binascii
import gzip
import hashlib
import io
import re
import tarfile

# pylint: disable=unused-import
from typing import Tuple, Any, Set, Optional, List, Iterable, Dict, Callable, Union
import zipfile
from collections import namedtuple

import pandas as pd

from ..common.utility import export
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


BinaryRecord = namedtuple(
    "BinaryRecord",
    [
        "reference",
        "original_string",
        "file_name",
        "file_type",
        "input_bytes",
        "decoded_string",
        "encoding_type",
        "file_hashes",
        "md5",
        "sha1",
        "sha256",
        "printable_bytes",
    ],
)

# pylint: disable=locally-disabled, line-too-long
_BASE64_HEADER_TYPES = {
    """TVqQAAMAAAAEAAAA//8AALgAAAAAAAAAQAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA+AAAAA4fug""": "exe",
    """TVqQAAMAAAAEAAAA//8AALgAAAAAAAAAQAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA8AAAAA4fug""": "dll",
    """TVqQAAMAAAAEAAAA//8AALgAAAAAAAAAQAAAAAAAAAAAAAAAAA\
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA6AAAAA4fug""": "sys",
    "UEsDBBQAAAAIA": "zip",
    "UEsDBBQAAQAIA": "zip (pwd protected)",
    "H4sI": "gz",
    "N3q8ryccAAR": "7z",
    "UmFyIRoHAM": "rar",
    "JVBERi0xLjcNC": "pdf",
    "0M8R4KGxGuE": "msi",
    "TVNXSU0AAADQ": "wim",
}
# pylint: enable=locally-disabled, line-too-long
_BASE64_HEADER_OFFSET_TYPES = {"DAxMDA3NzcAMDAwMDAwM": "tar"}

# Base64 simple regex
_BASE64_REGEX = "(?P<b64>[A-Za-z0-9+/\\n\\r]{30,}={0,2})"
BASE64_REGEX_C = re.compile(_BASE64_REGEX, re.I | re.X)
# Same expresion without group for pandas
_BASE64_REGEX_NG = "[A-Za-z0-9+/\\n\\r]{30,}={0,2}"

# we use this to store a set of strings that match the B64 regex but
# that we were unable to decode - so that we don't end up in an
# infinite loop
_UNDECODABLE_STRINGS: Set[str] = set()

# When True prints see more verbose execution
# (set from 'trace' parameter to unpack_items)
# pylint: disable=locally-disabled, invalid-name
_debug_trace = False
# pylint: enable=locally-disabled, invalid-name

_STRIP_TAGS = r"</?decoded[^>]*>"


def _get_trace_setting() -> Callable[[Optional[bool]], bool]:
    """Closure for holding trace setting."""
    _trace = False

    def _trace_enabled(trace: Optional[bool] = None) -> bool:
        nonlocal _trace
        if trace is not None:
            _trace = trace
        return _trace

    return _trace_enabled


_GET_TRACE = _get_trace_setting()


def _get_utf16_setting() -> Callable[[Optional[bool]], bool]:
    """Closure for holding utf16 decoding setting."""
    _utf16 = False

    def _utf16_enabled(utf16: Optional[bool] = None) -> bool:
        nonlocal _utf16
        if utf16 is not None:
            _utf16 = utf16
        return _utf16

    return _utf16_enabled


_GET_UTF16 = _get_utf16_setting()


@export
def unpack_items(
    input_string: str = None,
    data: pd.DataFrame = None,
    column: str = None,
    trace: bool = False,
    utf16: bool = False,
) -> Any:
    """
    Base64 decode an input string or strings taken from a pandas dataframe.

    Parameters
    ----------
    input_string : str, optional
        single string to decode (the default is None)
    data : pd.DataFrame, optional
        dataframe containing column to decode (the default is None)
    column : str, optional
        Name of dataframe text column (the default is None)
    trace : bool, optional
        Show additional status (the default is None)
    utf16 : bool, optional
        Attempt to decode UTF16 byte strings

    Returns
    -------
    Tuple[str, pd.DataFrame] (if `input_string`)
        Decoded string and additional metadata
    pd.DataFrame
        Decoded stringa and additional metadata in dataframe

    Notes
    -----
    If the input is a dataframe you must supply the name of the column to use.

    Items that decode to utf-8 or utf-16 strings will be returned as decoded
    strings replaced in the original string. If the encoded string is a
    known binary type it will identify the file type and return the hashes
    of the file. If any binary types are known archives (zip, tar, gzip) it
    will unpack the contents of the archive.
    For any binary it will return the decoded file as a byte array, and as a
    printable list of byte values. If the input is a string the function
    returns:

    - decoded string: this is the input string with any decoded sections
      replaced by the results of the decoding

    It also returns the data as a Pandas DataFrame with the following columns:

    - reference : this is an index that matches an index number in the
      returned string (e.g. <<encoded binary type=pdf index=1.2').
    - original_string : the string prior to decoding - file_type : the type
      of file if this could be determined
    - file_hashes : a dictionary of hashes (the md5, sha1 and sha256 hashes
      are broken out into separate columns)
    - input_bytes : the binary image as a byte array
    - decoded_string : printable form of the decoded string (either string
      or list of hex byte values)
    - encoding_type : utf-8, utf-16 or binary
    - md5, sha1, sha256 : the respective hashes of the binary file_type,
      file_hashes, input_bytes, md5, sha1, sha256 will be null if this item is
      decoded to a string

    If the input is a dataframe the output dataframe will also include the
    following column: - src_index - the index of the source row in the input
    frame. This allows you to re-join the output data to the input data.

    """
    _GET_TRACE(trace)
    _GET_UTF16(utf16)

    if input_string is not None:
        input_string = _b64_string_pad(input_string)
        return _decode_b64_string_recursive(input_string)
    if data is not None:
        if not column:
            raise ValueError("column must be supplied if the input is a DataFrame")
        return unpack_df(data=data, column=column, trace=trace, utf16=utf16)
    return None


def unpack(
    input_string: str, trace: bool = False, utf16: bool = False
) -> Tuple[str, Optional[List[BinaryRecord]]]:
    """
    Base64 decode an input string.

    Parameters
    ----------
    input_string : str, optional
        single string to decode (the default is None)
    trace : bool, optional
        Show additional status (the default is None)
    utf16 : bool, optional
        Attempt to decode UTF16 byte strings

    Returns
    -------
    Tuple[str, Optional[List[BinaryRecord]]]
        Decoded string and additional metadata

    Notes
    -----
    Items that decode to utf-8 or utf-16 strings will be returned as decoded
    strings replaced in the original string. If the encoded string is a
    known binary type it will identify the file type and return the hashes
    of the file. If any binary types are known archives (zip, tar, gzip) it
    will unpack the contents of the archive.
    For any binary it will return the decoded file as a byte array, and as a
    printable list of byte values. If the input is a string the function
    returns:

    - decoded string: this is the input string with any decoded sections
      replaced by the results of the decoding

    """
    _GET_TRACE(trace)
    _GET_UTF16(utf16)

    return _decode_b64_string_recursive(input_string)


def unpack_df(
    data: pd.DataFrame, column: str, trace: bool = False, utf16: bool = False
) -> pd.DataFrame:
    """
    Base64 decode strings taken from a pandas dataframe.

    Parameters
    ----------
    data : pd.DataFrame
        dataframe containing column to decode
    column : str
        Name of dataframe text column
    trace : bool, optional
        Show additional status (the default is None)
    utf16 : bool, optional
        Attempt to decode UTF16 byte strings

    Returns
    -------
    pd.DataFrame
        Decoded string and additional metadata in dataframe

    Notes
    -----
    Items that decode to utf-8 or utf-16 strings will be returned as decoded
    strings replaced in the original string. If the encoded string is a
    known binary type it will identify the file type and return the hashes
    of the file. If any binary types are known archives (zip, tar, gzip) it
    will unpack the contents of the archive.
    For any binary it will return the decoded file as a byte array, and as a
    printable list of byte values.

    The columns of the output DataFrame are:

    - decoded string: this is the input string with any decoded sections
      replaced by the results of the decoding
    - reference : this is an index that matches an index number in the
      decoded string (e.g. <<encoded binary type=pdf index=1.2').
    - original_string : the string prior to decoding
    - file_type : the type of file if this could be determined
    - file_hashes : a dictionary of hashes (the md5, sha1 and sha256 hashes
      are broken out into separate columns)
    - input_bytes : the binary image as a byte array
    - decoded_string : printable form of the decoded string (either string
      or list of hex byte values)
    - encoding_type : utf-8, utf-16 or binary
    - md5, sha1, sha256 : the respective hashes of the binary file_type,
      file_hashes, input_bytes, md5, sha1, sha256 will be null if this item is
      decoded to a string
    - src_index - the index of the source row in the input
      frame.

    """
    _GET_TRACE(trace)
    _GET_UTF16(utf16)

    output_df = pd.DataFrame(columns=BinaryRecord._fields)
    row_results: List[pd.DataFrame] = []
    rows_with_b64_match = data[data[column].str.contains(_BASE64_REGEX_NG)]
    for input_row in rows_with_b64_match[[column]].itertuples():
        (decoded_string, output_frame) = _decode_b64_string_recursive(input_row[1])
        output_frame["src_index"] = input_row.Index
        output_frame[column] = input_row._asdict()[column]
        output_frame["full_decoded_string"] = decoded_string
        row_results.append(output_frame)

    if row_results:
        output_df = pd.concat(row_results, ignore_index=True)
    return output_df


# pylint: disable=too-many-locals
def _decode_b64_string_recursive(
    input_string: str,
    max_recursion: int = 20,
    current_depth: int = 1,
    item_prefix: str = "",
) -> Tuple[str, pd.DataFrame]:
    """Recursively decode and unpack an encoded string."""
    _debug_print_trace("_decode_b64_string_recursive: ", max_recursion)
    _debug_print_trace("processing input: ", input_string[:200])

    decoded_string = input_string

    df_results = pd.DataFrame(columns=BinaryRecord._fields)
    fragment_index = 0
    match_pos = 0
    decode_success = False

    while True:
        # search sequentially through the input string for any strings
        # that look like base64
        _debug_print_trace(
            f"regex searching {decoded_string[:200]} ",
            f"from pos: {match_pos} bin_index {fragment_index}",
        )
        b64match = BASE64_REGEX_C.search(decoded_string, match_pos)
        if b64match is None:
            break

        b64_candidate = b64match.groupdict()["b64"]
        _debug_print_trace("regex found: ", b64_candidate)
        # if we already know that this string won't decode, skip
        if b64_candidate in _UNDECODABLE_STRINGS:
            match_pos = b64match.end()
            continue

        # try to decode
        fragment_index += 1
        (decoded_fragment, binary_items) = _decode_and_format_b64_string(
            b64_candidate,
            item_prefix=item_prefix,
            current_index=fragment_index,
            current_depth=current_depth,
        )

        decode_success = decoded_fragment != b64_candidate
        if decode_success:
            # we did decode something so lets put our result this in the output string
            if binary_items:
                new_records = _add_to_results(
                    binary_items,
                    b64_candidate,
                    current_depth,
                    item_prefix,
                    fragment_index,
                )
                df_results = df_results.append(
                    new_records, ignore_index=True, sort=False
                )
            # replace the decoded fragment in our current results string
            # (decode_string)
            decoded_string = decoded_string.replace(b64_candidate, decoded_fragment)
            _debug_print_trace(
                "Replaced string",
                decoded_string[match_pos : match_pos + 100],  # noqa: E203
            )
            match_pos += len(decoded_fragment)
        else:
            # if the string didn't decode we'll have the same output as input
            # so add that to our set of undecodable strings (we need to track this
            # otherwise we will recurse infinitely)
            _UNDECODABLE_STRINGS.add(b64_candidate)
            _debug_print_trace("new undecodable string")
            match_pos = b64match.end()

        if fragment_index > 50:
            break

    # if we reach our max recursion depth bail out here
    if max_recursion == 0:
        _debug_print_trace("max recursion reached")
        return decoded_string, df_results

    if decode_success:
        # stuff that we have already decoded may also contain further
        # base64 encoded strings
        prefix = (
            f"{item_prefix}.{fragment_index}." if item_prefix else f"{fragment_index}."
        )
        next_level_string, child_records = _decode_b64_string_recursive(
            decoded_string,
            item_prefix=prefix,
            max_recursion=max_recursion - 1,
            current_depth=(current_depth + 1),
        )
        return (
            next_level_string,
            df_results.append(child_records, ignore_index=True, sort=False),
        )

    _debug_print_trace("Nothing left to decode")
    return decoded_string, df_results


def _add_to_results(
    binary_items: Iterable[BinaryRecord],
    original_str: str,
    current_depth: int,
    item_prefix: str,
    fragment_index: int,
) -> List[Dict[str, Any]]:
    """Add current set of decoding results to collection."""
    new_rows = []
    for bin_record in binary_items:
        new_row = bin_record._asdict()
        new_row["reference"] = (
            f"{item_prefix}",
            f"{current_depth}.",
            f"{fragment_index}",
        )
        new_row["original_string"] = original_str
        new_row["md5"] = new_row["file_hashes"]["md5"]
        new_row["sha1"] = new_row["file_hashes"]["sha1"]
        new_row["sha256"] = new_row["file_hashes"]["sha256"]

        new_rows.append(new_row)
    return new_rows


def _debug_print_trace(*args):
    if _GET_TRACE():
        for arg in args:
            print(arg, end="")
        print()


# pylint: disable=too-many-locals
def _decode_and_format_b64_string(
    b64encoded_string: str,
    item_prefix: str = "",
    current_depth: int = 1,
    current_index: int = 1,
) -> Tuple[str, Optional[List[BinaryRecord]]]:
    """Decode string and return displayable content plus list of decoded artifacts."""
    # Check if we recognize this as a known file type
    (_, f_type) = _is_known_b64_prefix(b64encoded_string)
    _debug_print_trace("Found type: ", f_type)
    output_files = _decode_b64_binary(b64encoded_string, f_type)
    if not output_files:
        return b64encoded_string, None

    if len(output_files) == 1:
        # get the first (only) item
        out_name, out_record = output_files.popitem()
        _debug_print_trace("_decode_b64_binary returned a single record")
        _debug_print_trace("record:", out_record)

        disp_string = _format_single_record(
            out_name, out_record, item_prefix, current_depth, str(current_index)
        )
        return disp_string, [out_record]

    # Build header display string
    display_header = (
        f"<decoded value='multiple binary' type='multiple' "
        f"index='{item_prefix}{current_index}' "
        f"depth='{current_depth}'>"
    )
    child_display_strings = []
    child_depth = current_depth + 1
    _debug_print_trace("_decode_b64_binary returned multiple records")

    # Build child display strings
    for child_index, (child_name, child_rec) in enumerate(
        output_files.items(), start=1
    ):
        _debug_print_trace("Child_decode: ", child_rec)
        child_index_string = f"{item_prefix}{current_index}.{child_index}"
        disp_string = _format_single_record(
            child_name, child_rec, item_prefix, child_depth, child_index_string
        )
        child_display_strings.append(disp_string)
    display_string = display_header + "".join(child_display_strings) + "</decoded>"
    return display_string, list(output_files.values())


def _format_single_record(
    out_name: str,
    out_record: BinaryRecord,
    item_prefix: str,
    current_depth: int,
    current_index: str,
) -> str:
    # Build display string
    # If a string, include the decoded item in the output
    if out_record.encoding_type in ["utf-8", "utf-16"]:
        display_string = (
            f"<decoded type='string' name='{out_name}' "
            f"index='{item_prefix}{current_index}' "
            f"depth='{current_depth}'>"
            f"{out_record.decoded_string}</decoded>"
        )
        return display_string

    # if a binary include printable bytes
    display_string = (
        f"<decoded value='binary'  name='{out_name}' "
        f"type='{out_record.file_type}' "
        f"index='{item_prefix}{current_index}' "
        f"depth='{current_depth}'>"
        f"{out_record.printable_bytes}</decoded>"
    )
    return display_string


def _print_bytes(bytes_array: bytes):
    """Print byte array as string or hex."""
    enc_results = _get_byte_encoding(bytes_array)
    if enc_results.encoding_type != "binary":
        print(enc_results.decoded_string)

    else:
        print("Could not decode bytes to string. Hashes:")
        print(get_hashes(_binary_to_bytesio(bytes_array)))
        print(enc_results.printable_bytes)


def _as_byte_string(bytes_array) -> str:
    return " ".join(f"{b:02x}" for b in bytes_array)


def _empty_binary_rec() -> BinaryRecord:
    return BinaryRecord(
        reference=None,
        original_string=None,
        decoded_string=None,
        encoding_type=None,
        file_name=None,
        file_type=None,
        input_bytes=None,
        file_hashes=None,
        md5=None,
        sha1=None,
        sha256=None,
        printable_bytes=None,
    )


def _get_byte_encoding(bytes_array: bytes) -> BinaryRecord:
    """
    Return encoding type and decoded result.

    Decoded result is list of hex bytes in the case that the decoded
    result is not a string
    """
    result_rec = _empty_binary_rec()
    printable_bytes = _as_byte_string(bytes_array)
    if _GET_UTF16():  # type: ignore
        try:
            # Difficult to tell the difference between a real unicode string
            # and a binary string that happens to decode to a utf-16 string.
            # So we don't do this unless instructed to
            decoded_string = bytes_array.decode("utf-16")
            return result_rec._replace(
                decoded_string=decoded_string,
                encoding_type="utf-16",
                printable_bytes=printable_bytes,
            )
        except UnicodeDecodeError:
            pass
    try:
        decoded_string = bytes_array.decode("utf-8")
        return result_rec._replace(
            decoded_string=decoded_string,
            encoding_type="utf-8",
            printable_bytes=printable_bytes,
        )
    except UnicodeDecodeError:
        pass

    return result_rec._replace(encoding_type="binary", printable_bytes=printable_bytes)


def _is_known_b64_prefix(
    input_string: str,
) -> Union[Tuple[str, str], Tuple[None, None]]:
    """If this is known file type return the prefix and file type."""
    first160chars = input_string[0:160].replace("\n", "").replace("\r", "")
    for prefix, file_type in _BASE64_HEADER_TYPES.items():
        if first160chars.startswith(prefix):
            return prefix, file_type

    for matching_string, file_type in _BASE64_HEADER_OFFSET_TYPES.items():
        regex_comp = re.compile(matching_string, re.I | re.X)
        if regex_comp.search(first160chars):
            return matching_string, file_type
    return None, None


def _decode_b64_binary(
    input_string: str, file_type: str = None
) -> Optional[Dict[str, BinaryRecord]]:
    """Examine input string for known binaries and decode and unpack."""
    if not file_type:
        (_, f_type) = _is_known_b64_prefix(input_string)
        file_type = f_type

    try:
        decoded_bytes = base64.b64decode(input_string)
        return _unpack_and_hash_b64_binary(decoded_bytes, file_type)
    except binascii.Error:
        # we couldn't decode
        _debug_print_trace("Binascii exception - trying to decode string")
        _debug_print_trace(input_string)
        return None


def _unpack_and_hash_b64_binary(
    input_bytes: bytes, file_type: str = None
) -> Optional[Dict[str, BinaryRecord]]:
    """
    If this is a known archive type extract the contents.

    Return a dictionary of (file_type :
        (file_name, file_hashes, input_bytes, decoded_string, encoding_type)
    """
    if not input_bytes:
        return None

    output_files = {}
    if file_type in ["zip", "gz", "tar"]:
        # if this is a known archive type - try to extract the contents
        (unpacked_type, file_items) = _get_items_from_archive(input_bytes, file_type)
        if unpacked_type != "unknown":
            for file_name, extracted_file in file_items.items():
                file_results = _get_hashes_and_printable_string(extracted_file)
                idx = f"[{unpacked_type}] Filename: {file_name}"

                # ToDo - the unpacked type here refers to the archive file type  # pylint: disable=fixme
                # so assigning this to file_type is not exactly the right thing
                # to do. In a future episode we'll try to determine the file type
                # using magic numbers.
                output_files[idx] = file_results._replace(
                    file_name=file_name,
                    file_type=unpacked_type,
                    input_bytes=extracted_file,
                )
                _debug_print_trace(
                    "_unpack_and_hash_b64_binary item (archive): ",
                    type(file_results.decoded_string),
                    file_results.decoded_string,
                )

    if not output_files:
        # if this wasn't a known archive type or we failed to unpack anything,
        # just get the hashes and return
        file_results = _get_hashes_and_printable_string(input_bytes)
        idx = f"[{file_type}]"
        output_files[idx] = file_results._replace(
            file_name="unknown", file_type=file_type, input_bytes=input_bytes
        )
        _debug_print_trace(
            "_unpack_and_hash_b64_binary item (other): ",
            type(file_results.decoded_string),
            file_results.decoded_string,
        )
    return output_files


def _get_hashes_and_printable_string(extracted_file: bytes) -> BinaryRecord:
    """
    Get the hashes, encoding type and printable form of binary.

    (either string or list of hex-encoded byte values)
    """
    file_details = _get_byte_encoding(extracted_file)
    file_hashes = get_hashes(extracted_file)
    return file_details._replace(file_hashes=file_hashes)


def _get_items_from_archive(
    binary: bytes, archive_type: str = "zip"
) -> Tuple[str, Dict[str, bytes]]:
    """Extract contained files from an archive type."""
    _debug_print_trace("_get_items_from_archive type: ", archive_type)
    if archive_type == "zip":
        return get_items_from_zip(binary)
    if archive_type == "gz":
        return get_items_from_gzip(binary)
    if archive_type == "tar":
        return get_items_from_tar(binary)
    return "unknown", {archive_type: binary}


@export
def get_items_from_gzip(binary: bytes) -> Tuple[str, Dict[str, bytes]]:
    """
    Return decompressed gzip contents.

    Parameters
    ----------
    binary : bytes
        byte array of gz file

    Returns
    -------
    Tuple[str, bytes]
        File type + decompressed file

    """
    archive_file = gzip.decompress(binary)
    return "gz", {"gzip_file": archive_file}


@export
def get_items_from_zip(binary: bytes) -> Tuple[str, Dict[str, bytes]]:
    """
    Return dictionary of zip contents.

    Parameters
    ----------
    binary : bytes
        byte array of zip file

    Returns
    -------
    Tuple[str, Dict[str, bytes]]
        Filetype + dictionary of file name + file content

    """
    file_obj = io.BytesIO(binary)
    with zipfile.ZipFile(file_obj, mode="r") as zip_archive:
        archive_dict = {}
        for item in zip_archive.namelist():
            archive_file = zip_archive.read(item)
            archive_dict[item] = archive_file
        return "zip", archive_dict


@export
def get_items_from_tar(binary: bytes) -> Tuple[str, Dict[str, bytes]]:
    """
    Return dictionary of tar file contents.

    Parameters
    ----------
    binary : bytes
        byte array of zip file

    Returns
    -------
    Tuple[str, Dict[str, bytes]]
        Filetype + dictionary of file name + file content

    """
    file_obj = io.BytesIO(binary)
    # Open tarfile
    with tarfile.open(mode="r", fileobj=file_obj) as tar:
        archive_dict: Dict[str, bytes] = {}
        # Iterate over every member
        for item in tar.getnames():
            tar_file = tar.extractfile(item)
            archive_dict[item] = tar_file.read() if tar_file else b""
        return "tar", archive_dict


@export
def get_hashes(binary: bytes) -> Dict[str, str]:
    """
    Return md5, sha1 and sha256 hashes of input byte string.

    Parameters
    ----------
    binary : bytes
        byte string of item to be hashed

    Returns
    -------
    Dict[str, str]
        dictionary of hash algorithm + hash value

    """
    hash_dict = {}
    for hash_type in ["md5", "sha1", "sha256"]:
        if hash_type == "md5":
            hash_alg = hashlib.md5()  # nosec
        elif hash_type == "sha1":
            hash_alg = hashlib.sha1()  # nosec
        else:
            hash_alg = hashlib.sha256()
        hash_alg.update(binary)
        hash_dict[hash_type] = hash_alg.hexdigest()
    return hash_dict


def _binary_to_bytesio(binary: Union[bytes, io.BytesIO]) -> memoryview:
    if isinstance(binary, io.BytesIO):
        return binary.getbuffer()
    return io.BytesIO(binary).getbuffer()


def _b64_string_pad(string: str) -> str:
    if len(string) % 4 == 0:
        return string

    string = string.rstrip("=")
    while len(string) % 4 != 0:
        string = string + "A"
    return string


# pylint: disable=too-few-public-methods
@pd.api.extensions.register_dataframe_accessor("mp_b64")
class B64ExtractAccessor:
    """Base64 Unpack pandas extension."""

    def __init__(self, pandas_obj):
        """Initialize the extension."""
        self._df = pandas_obj

    def extract(self, column, **kwargs) -> pd.DataFrame:
        """
        Base64 decode strings taken from a pandas dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            dataframe containing column to decode
        column : str
            Name of dataframe text column
        trace : bool, optional
            Show additional status (the default is None)
        utf16 : bool, optional
            Attempt to decode UTF16 byte strings

        Returns
        -------
        pd.DataFrame
            Decoded string and additional metadata in dataframe

        Notes
        -----
        Items that decode to utf-8 or utf-16 strings will be returned as decoded
        strings replaced in the original string. If the encoded string is a
        known binary type it will identify the file type and return the hashes
        of the file. If any binary types are known archives (zip, tar, gzip) it
        will unpack the contents of the archive.
        For any binary it will return the decoded file as a byte array, and as a
        printable list of byte values.

        The columns of the output DataFrame are:

        - decoded string: this is the input string with any decoded sections
          replaced by the results of the decoding
        - reference : this is an index that matches an index number in the
          decoded string (e.g. <<encoded binary type=pdf index=1.2').
        - original_string : the string prior to decoding - file_type : the type
          of file if this could be determined
        - file_hashes : a dictionary of hashes (the md5, sha1 and sha256 hashes
          are broken out into separate columns)
        - input_bytes : the binary image as a byte array
        - decoded_string : printable form of the decoded string (either string
          or list of hex byte values)
        - encoding_type : utf-8, utf-16 or binary
        - md5, sha1, sha256 : the respective hashes of the binary file_type,
          file_hashes, input_bytes, md5, sha1, sha256 will be null if this item is
          decoded to a string
        - src_index - the index of the source row in the input
          frame.

        """
        return unpack_df(data=self._df, column=column, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Azure authentication handling."""
import os
from typing import List

from azure.common.exceptions import CloudError
from azure.mgmt.subscription import SubscriptionClient

from .._version import VERSION

# importing only_interactive_cred for client use.
# pylint: disable=unused-import
from .azure_auth_core import (  # noqa: F401
    AzCredentials,
    az_connect_core,
    AzureCloudConfig,
    only_interactive_cred,
)

# pylint: enable=unused-import
from .provider_settings import get_provider_settings

__version__ = VERSION
__author__ = "Pete Bryan"


def az_connect(
    auth_methods: List[str] = None,
    silent: bool = False,
) -> AzCredentials:
    """
    Connect to Azure SDK/API.

    Parameters
    ----------
    auth_methods : List[str], optional
        List of authentication methods to try
        Possible options are:
        - "env" - to get authentication details from environment variables
        - "cli" - to use Azure CLI authentication details
        - "msi" - to user Managed Service Identity details
        - "interactive" - to prompt for interactive login
        Default is ["env", "cli", "msi", "interactive"]
    silent : bool, optional
        Set True to hide all output during connection, by default False

    Returns
    -------
    AzCredentials
        Named tuple of:
        - legacy (ADAL) credentials
        - modern (MSAL) credentials


    Raises
    ------
    CloudError
        If chained token credential creation fails.

    """
    az_cloud_config = AzureCloudConfig()
    # If using env options try to load from msticpy
    data_provs = get_provider_settings(config_section="DataProviders")
    az_cli_config = data_provs.get("AzureCLI")
    auth_methods = auth_methods or az_cloud_config.auth_methods
    if az_cli_config and az_cli_config.args:
        if "auth_methods" in az_cli_config.args:
            auth_methods = az_cli_config.args.get("auth_methods")
        if isinstance(auth_methods, list) and "env" in auth_methods:
            os.environ["AZURE_CLIENT_ID"] = az_cli_config.args.get("clientId") or ""
            os.environ["AZURE_TENANT_ID"] = az_cli_config.args.get("tenantId") or ""
            os.environ["AZURE_CLIENT_SECRET"] = (
                az_cli_config.args.get("clientSecret") or ""
            )
    credentials = az_connect_core(auth_methods=auth_methods, silent=silent)
    sub_client = SubscriptionClient(
        credential=credentials.modern,
        base_url=az_cloud_config.endpoints.resource_manager,
        credential_scopes=[az_cloud_config.token_uri],
    )
    if not sub_client:
        raise CloudError("Could not create a Subscription client.")

    return credentials


def az_user_connect(silent: bool = False) -> AzCredentials:
    """
    Authenticate to the SDK using user based authentication methods, Azure CLI or interactive logon.

    Parameters
    ----------
    silent : bool, optional
        Whether you want the auth process to display any output, by default False

    Returns
    -------
    AzCredentials

    """
    return az_connect_core(auth_methods=["cli", "interactive"], silent=silent)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Text to DataFrame IPython magic."""
from IPython.core.magic import needs_local_scope, register_cell_magic
from .pivot_magic_core import run_txt2df

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@register_cell_magic
@needs_local_scope
def txt2df(line, cell, local_ns):
    """Convert cell text to pandas DataFrame."""
    return run_txt2df(line, cell, local_ns)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pandas DataFrame accessor for Pivot functions."""
import re
import json
from json import JSONDecodeError
from datetime import datetime
from numbers import Number
import warnings
from typing import Callable, Dict, Iterable, Set, Union

import numpy as np
import pandas as pd
from IPython import get_ipython
from IPython.display import HTML, display

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def _verbose_out(data, func, st_time, verbose=False, debug=False, **kwargs):
    """Output verbose or debug info."""
    end_time = datetime.now()

    if verbose or debug:
        print(f"{len(data)} rows returned from function {func.__name__}")
    if debug:
        print(f"Columns in result from function {func.__name__}:")
        print(data.columns)
        print(f"Execution time: {end_time - st_time}")
        print(f"Parameters: {kwargs}")


@pd.api.extensions.register_dataframe_accessor("mp_pivot")
class PivotAccessor:
    """Pandas api extension for Pivot functions."""

    def __init__(self, pandas_obj):
        """Instantiate pivot extension class."""
        self._df = pandas_obj
        self._ip = get_ipython()

    def run(self, func: Callable[..., pd.DataFrame], **kwargs) -> pd.DataFrame:
        """
        Run a pivot function on the current DataFrame.

        Parameters
        ----------
        func : Callable[..., pd.DataFrame]
            Pivot function to run
        kwargs:
            Keyword arguments to pass to `func`.
            A column specification (e.g. column="src_col_name")
            is usually the minimum needed.
            For data queries the `column` keyword must be the
            name of the the query parameter (e.g. host_name = "src_col_name")

        Returns
        -------
        pd.DataFrame
            The output DataFrame from the function.

        Notes
        -----
        You can pass the `join` keyword argument to most
        pivot functions. Values for join are "inner", "left", "right"
        or "outer".

        """
        st_time = datetime.now()
        verbose = kwargs.pop("verbose", False)
        debug = kwargs.pop("debug", False)
        result = self._df.pipe((func, "data"), **kwargs)
        if verbose or debug:
            _verbose_out(
                data=result,
                func=func,
                st_time=st_time,
                verbose=verbose,
                debug=debug,
                **kwargs,
            )
        return result

    def display(
        self,
        title: str = None,
        cols: Iterable[str] = None,
        query: str = None,
        head: int = None,
    ) -> pd.DataFrame:
        """
        Display the DataFrame in the middle of a pipeline.

        Parameters
        ----------
        title : str, optional
            Title to display for the DataFrame, by default None
        cols : Iterable[str], optional
            List of columns to display, by default None
        query : str, optional
            Query to filter the displayed data, by default None
            This should be a string executable by the DataFrame.query
            function
        head : int, optional
            Limit the displayed output to `head` rows, by default None

        Returns
        -------
        pd.DataFrame
            Passed through input DataFrame.

        """
        if title:
            display(HTML(f"<h3>{title}</h3>"))
        disp_df = self._df
        if cols:
            disp_df = disp_df[cols]
        if query:
            disp_df = disp_df.query(query)  # , parser='python', engine='python')
        if head:
            disp_df = disp_df.head(head)
        display(disp_df)
        return self._df

    def tee(self, var_name: str, clobber: bool = False) -> pd.DataFrame:
        """
        Save current dataframe to `var_name` in the IPython user namespace.

        Parameters
        ----------
        var_name : str
            The name of the DF variable to create.
        clobber : bool, optional
            Whether to overwrite an existing variable of the same name,
            by default False

        Returns
        -------
        pd.DataFrame
            Passed through input DataFrame.

        Notes
        -----
        This function only works in an IPython/Jupyter notebook environment.
        It will attempt to create a variable in the user local namespace
        that references the current state of the DataFrame in the pipeline.

        By default it will not overwrite an existing variable of the same
        name (specify `clobber=True` to overwrite)

        """
        if self._ip and var_name:
            if var_name in self._ip.ns_table["user_local"] and not clobber:
                warnings.warn(f"Did not overwrite existing {var_name} in namespace")
            else:
                self._ip.ns_table["user_local"][var_name] = self._df
        return self._df

    def tee_exec(self, df_func: str, *args, **kwargs) -> pd.DataFrame:
        """
        Run a dataframe method on the dataframe without changing it.

        Parameters
        ----------
        df_func : str
            The name of the function to execute. Accessor methods
            must be of the form "accessor.method".
        args : tuple
            Positional arguments to be passed to the function
        kwargs : dict
            Keyword arguments to be passed to the function.

        Returns
        -------
        pd.DataFrame
            Passed through input DataFrame.

        Notes
        -----
        This function runs the DataFrame method or accessor function.
        It does not alter the DataFrame (unless the function does
        any kind of in-place modification). The function is run and
        the original input DataFrame is returned.

        """
        acc_name = func_name = func = None
        if "." in df_func:
            acc_name, func_name = df_func.split(".")
            accessor = getattr(self._df, acc_name, None)
            if accessor:
                func = getattr(accessor, func_name, None)
        else:
            func = getattr(self._df, df_func, None)
        if func:
            # run the function with any additional args
            func(*args, **kwargs)
        return self._df

    def filter_cols(
        self,
        cols: Union[str, Iterable[str]],
        match_case: bool = False,
        sort_cols: bool = False,
    ) -> pd.DataFrame:
        """
        Filter output columns matching names in `cols` expression(s).

        Parameters
        ----------
        cols : Union[str, Iterable[str]]
            Either a string or a list of strings with filter expressions.
            These can be exact matches for column names, wildcard patterns
            ("*" matches multiple chars and "?" matches a single char),
            or regular expressions.
        match_case: bool, optional
            Use case-sensitive matching, by default False
        sort_cols : bool, optional
            Alphabetically sort column names, by default False

        Returns
        -------
        pd.DataFrame
            The input DataFrame with only columns that match the
            filtering expressions.

        """
        curr_cols = self._df.columns
        filt_cols: Set[str] = set()
        if isinstance(cols, str):
            filt_cols.update(_name_match(curr_cols, cols, match_case))
        elif isinstance(cols, list):
            for col_filter in cols:
                filt_cols.update(_name_match(curr_cols, col_filter, match_case))
        if not filt_cols:
            raise ValueError("Column filter expression(s) did not match any columns")

        if sort_cols:
            out_cols = sorted(filt_cols)
        else:
            # keep the existing order
            out_cols = [col for col in curr_cols if col in filt_cols]
        return self._df[out_cols]

    def filter(
        self,
        expr: Union[str, Number],
        match_case: bool = False,
        numeric_col: bool = False,
    ) -> pd.DataFrame:
        """
        Filter all columns of DataFrame, return rows with any matches.

        Parameters
        ----------
        expr : Union[str, Number]
            String or regular expression to match or a (partial) number.
            If `expr` is a string it is matched against any string or object
            columns using pandas `str.contains(..regex=True)`
            If `expr` is a number or if `numeric_col` is True, `expr`
            is converted to a string and matched as a substring of any numeric
            columns.
        match_case : bool, optional
            The match is not case-sensitive by default. Set to True to force
            case-sensitive matches.
        numeric_col : bool, optional
            If `expr` is a numeric string or number this will force a match against
            only numeric columns, by default False

        Returns
        -------
        pd.DataFrame
            The filtered dataframe

        Raises
        ------
        TypeError
            If `expr` is neither a string or number.

        """
        if isinstance(expr, str) and not numeric_col:
            text_cols = self._df.select_dtypes(include=[object, "string"])
            return self._df[
                text_cols.apply(
                    lambda col: col.str.contains(
                        expr, regex=True, case=match_case, na=False
                    )
                ).any(axis=1)
            ]
        if isinstance(expr, Number) or numeric_col:
            num_cols = self._df.select_dtypes(include="number")
            str_expr = str(expr)
            return self._df[
                num_cols.apply(
                    lambda col: col.astype("string").str.contains(str_expr, regex=True)
                ).any(axis=1)
            ]
        raise TypeError("expr '{expr}' must be a string or numeric type.")

    def sort(
        self, cols: Union[str, Iterable[str], Dict[str, str]], ascending: bool = None
    ) -> pd.DataFrame:
        """
        Sort output by column expression.

        Parameters
        ----------
        cols : Union[str, Iterable[str], Dict[str, str]]
            If this is a string, then this should be a column name expression. A column name
            expression is either a column name, a case-insenstive column name or a
            regular expression to match one or more column names.
            Each column name expression can be of
            the format `col_name_expr:desc` to sort descending (`col_name_expr:asc` is the default).
            The col_name can also be a regular expression or partial column name.
            If this is a list, then each element should be a column name expression
            with an optional ':asc' or ':desc' suffix.
            If this is a dict, then the keys should be column name expressions and the
            values bools indication 'ascending' (True) or 'descending' (False) sort.
        ascending : [type], optional
            Overrides any ordering specified for individual columns and sorts
            'ascending' if True or 'descending' if False. If not supplied and no
            column-specific ordering is supplied it sorts ascending.

        Returns
        -------
        pd.DataFrame
            The sorted DataFrame

        Raises
        ------
        ValueError
            One or more column expressions matched no column name in the input.

        """
        if isinstance(cols, dict):
            col_dict = cols
        else:
            if isinstance(cols, str):
                col_list = [col.strip() for col in cols.split(",")]
            else:
                col_list = list(cols)
            col_dict = {
                col.split(":")[0].strip(): not col.casefold().endswith(":desc")
                for col in col_list
            }

        sort_cols = {}
        # create case-insensitive mapping for DF cols
        df_cols = {col.casefold(): col for col in self._df.columns}
        for col in col_dict:
            # Use case-matched name, if available
            if col in self._df.columns:
                sort_cols[col] = col_dict[col]
                continue
            # Look for case-insensitive match
            df_col = df_cols.get(col.casefold())
            if df_col:
                sort_cols[col] = col_dict[col]
                continue
            # look for regex matches for col name
            df_match_cols = [
                df_cols[s_col]
                for s_col in df_cols
                if re.match(col, s_col, re.IGNORECASE)
            ]
            # we might get multiple matches
            if df_match_cols:
                sort_cols.update({df_col: col_dict[col] for df_col in df_match_cols})
                continue
            raise ValueError(
                f"'{col}' column in sort list did not match any columns in input data."
            )
        # create the ascending parameter
        asc_param = ascending if ascending is not None else list(sort_cols.values())
        return self._df.sort_values(list(sort_cols.keys()), ascending=asc_param)

    def list_to_rows(self, cols: Union[str, Iterable[str]]) -> pd.DataFrame:
        """
        Expand a list column to individual rows.

        Parameters
        ----------
        cols : Union[str, Iterable[str]]
            The columns to be expanded.

        Returns
        -------
        pd.DataFrame
            The expanded DataFrame

        """
        orig_cols = self._df.columns
        data = self._df
        if isinstance(cols, str):
            cols = [cols]
        for col in cols:
            item_col = f"{col}_list_item$$"
            ren_col = {item_col: col}
            data = (
                pd.DataFrame(data[col].to_list(), index=data.index)
                .replace([None], np.nan)  # convert any Nones to NaN
                .merge(data, right_index=True, left_index=True)
                .melt(id_vars=orig_cols, value_name=item_col)
                .dropna(subset=[item_col])  # get rid of rows with NaNs in this col
                .drop([col, "variable"], axis=1)
                .rename(columns=ren_col)
            )
        return data

    def parse_json(self, cols: Union[str, Iterable[str]]) -> pd.DataFrame:
        """
        Convert JSON string columns to Python types.

        Parameters
        ----------
        cols : Union[str, Iterable[str]]
            Column or interable of columns to process

        Returns
        -------
        pd.DataFrame
            Processed dataframe

        """
        if isinstance(cols, str):
            cols = [cols]
        data = self._df
        for col in cols:
            col_parsed = f"{col}_parsed"
            data[col_parsed] = data[col].apply(_json_safe_conv)
            data = data.drop([col], axis=1).rename(columns={col_parsed: col})
        return data


def _name_match(cur_cols: Iterable[str], col_filter, match_case):
    col_filter = re.sub(r"[^.]\*", ".*", col_filter)
    col_filter = re.sub(r"[^.]\?", ".?", col_filter)
    regex_opts = [re.IGNORECASE] if not match_case else []
    return {col for col in cur_cols if re.match(col_filter, col, *regex_opts)}


def _json_safe_conv(val):
    if val:
        try:
            return json.loads(val)
        except (TypeError, JSONDecodeError):
            pass
    return val



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Functions to support investigation of a domain or url.

Includes functions to conduct common investigation steps when dealing
with a domain or url, such as getting a screenshot or validating the TLD.

"""
from datetime import datetime
from enum import Enum
import json
import ssl
import time
from typing import Any, Dict, Optional, Tuple
from urllib.error import HTTPError, URLError

import cryptography as crypto
from cryptography.x509 import Certificate
import pandas as pd
import requests
import tldextract

# pylint: disable=no-name-in-module
from dns.resolver import Resolver
from dns.exception import DNSException

# pylint: enable=no-name-in-module
from IPython import display
from ipywidgets import IntProgress
from urllib3.exceptions import LocationParseError
from urllib3.util import parse_url

from .._version import VERSION
from ..common import pkg_config as config
from ..common.exceptions import MsticpyUserConfigError
from ..common.utility import export

__version__ = VERSION
__author__ = "Pete Bryan"


@export
def screenshot(url: str, api_key: str = None) -> requests.models.Response:
    """
    Get a screenshot of a url with Browshot.

    Parameters
    ----------
    url : str
        The url a screenshot is wanted for.
    api_key : str (optional)
        Browshot API key. If not set msticpyconfig checked for this.

    Returns
    -------
    image_data: requests.models.Response
        The final screenshot request response data.

    """
    # Get Browshot API key from kwargs or config
    if api_key is not None:
        bs_api_key: Optional[str] = api_key
    else:
        bs_conf = config.settings.get("DataProviders", {}).get(
            "Browshot"
        ) or config.settings.get("Browshot")
        bs_api_key = None
        if bs_conf is not None:
            bs_api_key = bs_conf.get("Args", {}).get("AuthKey")  # type: ignore

    if bs_api_key is None:
        raise MsticpyUserConfigError(
            "No configuration found for Browshot",
            "Please add a section to msticpyconfig.yaml:",
            "DataProviders:",
            "  Browshot:",
            "    Args:",
            "      AuthKey: {your_auth_key}",
            title="Browshot configuration not found",
            browshot_uri=("Get an API key for Browshot", "https://api.browshot.com/"),
        )

    # Request screenshot from Browshot and get request ID
    id_string = f"https://api.browshot.com/api/v1/screenshot/create?url={url}/&instance_id=26&size=screen&cache=0&key={bs_api_key}"  # pylint: disable=line-too-long
    id_data = requests.get(id_string)
    bs_id = json.loads(id_data.content)["id"]
    status_string = (
        f"https://api.browshot.com/api/v1/screenshot/info?id={bs_id}&key={bs_api_key}"
    )
    image_string = f"https://api.browshot.com/api/v1/screenshot/thumbnail?id={bs_id}&zoom=50&key={bs_api_key}"  # pylint: disable=line-too-long
    # Wait until the screenshot is ready and keep user updated with progress
    print("Getting screenshot")
    progress = IntProgress(min=0, max=40)
    display.display(progress)
    ready = False
    while not ready:
        progress.value += 1
        status_data = requests.get(status_string)
        status = json.loads(status_data.content)["status"]
        if status == "finished":
            ready = True
        else:
            time.sleep(0.05)
    progress.value = 40

    # Once ready get the screenshot
    image_data = requests.get(image_string)

    if image_data.status_code != 200:
        print(
            "There was a problem with the request, please check the status code for details"
        )

    return image_data


# Backward compat with dnspython 1.x
# If v2.x installed use non-deprecated "resolve" method
# otherwise use "query"
_dns_resolver = Resolver()
if hasattr(_dns_resolver, "resolve"):
    _dns_resolve = getattr(_dns_resolver, "resolve")
else:
    _dns_resolve = getattr(_dns_resolver, "query")


@export
class DomainValidator:
    """Assess a domain's validity."""

    _ssl_abuse_list: pd.DataFrame = pd.DataFrame()

    @classmethod
    def _check_and_load_abuselist(cls):
        """Pull IANA TLD list and save to internal attribute."""
        if cls._ssl_abuse_list is None or cls._ssl_abuse_list.empty:
            cls._ssl_abuse_list: pd.DataFrame = cls._get_ssl_abuselist()

    @property
    def ssl_abuse_list(self) -> pd.DataFrame:
        """
        Return the class SSL Blacklist.

        Returns
        -------
        pd.DataFrame
            SSL Blacklist

        """
        self._check_and_load_abuselist()
        return self._ssl_abuse_list

    @staticmethod
    def validate_tld(url_domain: str) -> bool:
        """
        Validate if a domain's TLD is valid.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        result:
            True if valid public TLD, False if not.

        """
        _, _, tld = tldextract.extract(url_domain.lower())
        return bool(tld)

    @staticmethod
    def is_resolvable(url_domain: str) -> bool:  # pylint: disable=no-self-use
        """
        Validate if a domain or URL be be resolved to an IP address.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        result:
            True if valid resolvable, False if not.

        """
        try:
            _dns_resolve(url_domain, "A")
            return True
        except DNSException:
            return False

    def in_abuse_list(self, url_domain: str) -> Tuple[bool, Optional[Certificate]]:
        """
        Validate if a domain or URL's SSL cert the abuse.ch SSL Abuse List.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        Tuple[bool, Optional[Certificate]]:
            True if valid in the list, False if not.
            Certificate - the certificate loaded from the domain.

        """
        x509: Optional[Certificate]
        try:
            cert = ssl.get_server_certificate((url_domain, 443))
            # pylint: disable=no-value-for-parameter
            x509 = crypto.x509.load_pem_x509_certificate(  # type: ignore
                cert.encode("ascii")
            )
            # pylint: enable=no-value-for-parameter
            cert_sha1 = x509.fingerprint(
                crypto.hazmat.primitives.hashes.SHA1()  # type: ignore # nosec
            )
            result = bool(
                self.ssl_abuse_list["SHA1"]
                .str.contains(cert_sha1.hex())
                .any()  # type: ignore
            )
        except Exception:  # pylint: disable=broad-except
            result = False
            x509 = None

        return result, x509

    @classmethod
    def _get_ssl_abuselist(cls) -> pd.DataFrame:
        """Download and load abuse.ch SSL Abuse List."""
        try:
            ssl_ab_list = pd.read_csv(
                "https://sslbl.abuse.ch/blacklist/sslblacklist.csv", skiprows=8
            )
        except (ConnectionError, HTTPError, URLError):
            ssl_ab_list = pd.DataFrame({"SHA1": []})

        return ssl_ab_list


def dns_components(domain: str) -> dict:
    """
    Return components of domain as dict.

    Parameters
    ----------
    domain : str
        The domain to extract.

    Returns
    -------
    dict:
        Returns subdomain and TLD components from a domain.

    """
    return tldextract.extract(domain.lower())._asdict()


def url_components(url: str) -> Dict[str, str]:
    """Return parsed Url components as dict."""
    try:
        return parse_url(url)._asdict()
    except LocationParseError:
        return {}


def dns_resolve(url_domain: str, rec_type: str = "A") -> Dict[str, Any]:
    """
    Validate if a domain or URL be be resolved to an IP address.

    Parameters
    ----------
    url_domain : str
        The url or domain to validate.
    rec_type : str
        The DNS record type to query, by default "A"

    Returns
    -------
    Dict[str, Any]:
        Resolver result as dictionary.

    """
    domain = parse_url(url_domain).host
    try:
        return _resolve_resp_to_dict(_dns_resolve(domain, rdtype=rec_type))
    except DNSException as err:
        return {
            "qname": domain,
            "rdtype": rec_type,
            "response": str(err),
        }


def ip_rev_resolve(ip_address: str) -> Dict[str, Any]:
    """
    Reverse lookup for IP Address.

    Parameters
    ----------
    ip_address : str
        The IP address to query.

    Returns
    -------
    Dict[str, Any]:
        Resolver result as dictionary.

    """
    try:
        return _resolve_resp_to_dict(_dns_resolve(ip_address, raise_on_no_answer=True))
    except DNSException as err:
        return {
            "qname": ip_address,
            "rdtype": "PTR",
            "response": str(err),
        }


def _resolve_resp_to_dict(resolver_resp):
    """Return Dns Python resolver response to dict."""
    rdtype = (
        resolver_resp.rdtype.name
        if isinstance(resolver_resp.rdtype, Enum)
        else str(resolver_resp.rdtype)
    )
    rdclass = (
        resolver_resp.rdclass.name
        if isinstance(resolver_resp.rdclass, Enum)
        else str(resolver_resp.rdclass)
    )

    return {
        "qname": str(resolver_resp.qname),
        "rdtype": rdtype,
        "rdclass": rdclass,
        "response": str(resolver_resp.response),
        "nameserver": getattr(resolver_resp, "nameserver", None),
        "port": getattr(resolver_resp, "port", None),
        "canonical_name": str(resolver_resp.canonical_name),
        "rrset": [str(res) for res in resolver_resp.rrset],
        "expiration": datetime.utcfromtimestamp(resolver_resp.expiration),
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process Entity class."""
from datetime import datetime
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .account import Account
from .entity_enums import ElevationToken
from .file import File
from .host import Host
from .host_logon_session import HostLogonSession

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name, too-many-instance-attributes


@export
class Process(Entity):
    """
    Process Entity class.

    Attributes
    ----------
    ProcessId : str
        Process ProcessId
    CommandLine : str
        Process CommandLine
    ElevationToken : str
        Process ElevationToken
    CreationTimeUtc : datetime
        Process CreationTimeUtc
    ImageFile : File
        Process ImageFile
    Account : Account
        Process Account
    ParentProcess : Process
        Process ParentProcess
    Host : Host
        Process Host
    LogonSession : HostLogonSession
        Process LogonSession

    """

    ID_PROPERTIES = ["ProcessId", "ImageFile", "CreationTimeUtc", "CommandLine"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        role="new",
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)
        role : str, optional
            'new' or 'parent' - only relevant if the entity
            is being constructed from an event.
            (the default is 'new')

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.ProcessId: Optional[str] = None
        self.CommandLine: Optional[str] = None
        self.ElevationToken: Optional[ElevationToken] = None
        self.CreationTimeUtc: datetime = datetime.min
        self.ImageFile: Optional[File] = None
        self.Account: Optional[Account] = None
        self.ParentProcess: Optional[Process] = None
        self.Host: Optional[Host] = None
        self.LogonSession: Optional[HostLogonSession] = None
        super().__init__(src_entity=src_entity, **kwargs)

        # pylint: disable=locally-disabled, line-too-long
        if src_event is not None:
            self._create_from_event(src_event, role)

    def _create_from_event(self, src_event, role):
        if role == "new":
            self.ProcessId = src_event.get("NewProcessId")
            self.CommandLine = src_event.get("CommandLine")
            if "TimeCreatedUtc" in src_event:
                self.CreationTimeUtc = src_event["TimeCreatedUtc"]
            elif "TimeGenerated" in src_event:
                self.CreationTimeUtc = src_event["TimeGenerated"]
            self.ImageFile = File(src_event=src_event, role="new")
            self.Account = Account(src_event=src_event, role="subject")

            if "ParentProcessName" in src_event or "ProcessName" in src_event:
                parent = Process(src_event=src_event, role="parent")
                self.ParentProcess = parent

            # Linux properties
            self.success = src_event.get("success")
            self.audit_user = src_event.get("audit_user")
            self.auid = src_event.get("auid")
            self.group = src_event.get("group")
            self.gid = src_event.get("gid")
            self.effective_user = src_event.get("effective_user")
            self.euid = src_event.get("euid")
            self.effective_group = src_event.get("effective_group")
            self.egid = src_event.get("effective_group")
            self.cwd = src_event.get("cwd")
            self.name = src_event.get("cwd")
        else:
            self.ProcessId = src_event.get("ProcessId")
            self.ImageFile = File(src_event=src_event, role="parent")

    @property
    def ProcessName(self) -> Optional[str]:  # noqa: N802
        """Return the name of the process file."""
        file = self["ImageFile"]
        return file.Name if file else None

    @property
    def ProcessFilePath(self) -> Optional[str]:  # noqa: N802
        """Return the name of the process file path."""
        file = self.ImageFile
        return file.FullPath if file else None

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        if self.ProcessFilePath:
            return f"{self.ProcessFilePath}: {self.CommandLine}"
        return self.__class__.__name__

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        if self.ImageFile:
            return f"{self.ImageFile.name_str}[pid:{self.ProcessId}]"
        return self.ImageFile.name_str if self.ImageFile else super().name_str

    _entity_schema = {
        # ProcessId (type System.String)
        "ProcessId": None,
        # CommandLine (type System.String)
        "CommandLine": None,
        # ElevationToken (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.ElevationToken])
        "ElevationToken": None,
        # CreationTimeUtc (type System.Nullable`1[System.DateTime])
        "CreationTimeUtc": None,
        # ImageFile (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.File)
        "ImageFile": "File",
        # Account (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Account)
        "Account": "Account",
        # ParentProcess (type Microsoft.Azure.Security.Detection.AlertContracts
        # .V3.Entities.Process)
        "ParentProcess": "Process",
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Host)
        "Host": "Host",
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.HostLogonSession)
        "LogonSession": "HostLogonSession",
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }
